from dataclasses import dataclass
from enum import auto
from typing import TYPE_CHECKING, Optional

from mstrio.modeling.schema.helpers import ObjectSubType, SchemaObjectReference
from mstrio.object_management import search_operations
from mstrio.object_management.search_enums import SearchPattern
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import (
    Dictable,
    camel_to_snake,
    delete_none_values,
    get_valid_project_id,
    snake_to_camel,
)

from .enums import DependenceType, DimtyType, ExpressionType, NodeType

if TYPE_CHECKING:
    from mstrio.connection import Connection


@dataclass(kw_only=True)
class ExpressionNode(Dictable):
    """Base class representing a node of an expression tree."""

    _TYPE = None
    _FROM_DICT_MAP = {
        'expression_type': ExpressionType,
        'dimty_type': DimtyType,
        'dependence_type': DependenceType,
    }

    expression_type: ExpressionType | None = None
    dimty_type: DimtyType | None = None
    dependence_type: DependenceType | None = None

    def to_dict(self, camel_case: bool = True) -> dict:
        result = super().to_dict(camel_case)
        result['type'] = self._TYPE.value

        return result

    @staticmethod
    def dispatch(
        source, connection: Optional['Connection'] = None
    ) -> type['ExpressionNode']:
        """Method dispatching node data to appropriate class that represents
        this type of node.
        """
        data = source.copy()
        node_type = NodeType(data.pop('type'))

        # this import is here due to circular imports issue
        from mstrio.modeling.expression.expression_nodes import NODE_TYPE_TO_CLASS_MAP

        cls = NODE_TYPE_TO_CLASS_MAP[node_type]

        return cls.from_dict(data, connection)


@dataclass
class Token(Dictable):
    """Class representation of a single tokens of an expression.

    Attributes:
        type: Enumeration constant that classifies the text within this token

        value: The raw text represented by a token

        target: If the token represents an object, provide information about
            the object

        level: Describe the amount of processing performed on this parser token

        state: Whether token is in an error or not

        attribute_form: If the token represents an attribute form in the context
           of an object (say City@DESC) then provide attribute form id
    """

    class Type(AutoName):
        """Enumeration constant that classifies the text within a token"""

        END_OF_TEXT = auto()
        ERROR = auto()
        UNKNOWN = auto()
        EMPTY = auto()
        CHARACTER = auto()
        LITERAL = auto()
        IDENTIFIER = auto()
        STRING_LITERAL = auto()
        INTEGER = auto()
        FLOAT = auto()
        BOOLEAN = auto()
        GUID = auto()
        OBJECT_REFERENCE = auto()
        COLUMN_REFERENCE = auto()
        OBJECT_AT_FORM = auto()
        FUNCTION = auto()
        KEYWORD = auto()
        OTHER = auto()
        ELEMENTS = auto()
        DATE_TIME = auto()

    class Level(AutoName):
        """Enumeration constant describing the amount of processing performed
        on a parser token
        """

        CLIENT = auto()
        LEXED = auto()
        RESOLVED = auto()
        PARSED = auto()

    class State(AutoName):
        """Enumeration constant describing whether token in an error or not"""

        ERROR = auto()
        INITIAL = auto()
        OKAY = auto()

    _FROM_DICT_MAP = {
        'type': Type,
        'target': SchemaObjectReference,
        'level': Level,
        'state': State,
    }

    value: str
    type: Type | None = None
    target: SchemaObjectReference | None = None
    attribute_form: str | None = None
    level: Level | None = None
    state: State | None = None

    def to_dict(self, camel_case=True) -> dict:
        result = super().to_dict(camel_case=camel_case)
        if self.attribute_form:
            result['attribute_form'] = {
                'object_id': self.attribute_form,
            }

        result = delete_none_values(result, recursion=False)
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(
        cls,
        source: dict,
        connection: Optional['Connection'] = None,
        to_snake_case: bool = True,
    ):
        data = camel_to_snake(source) if to_snake_case else source.copy()

        attribute_form = data.get('attribute_form')
        data['attribute_form'] = attribute_form['object_id'] if attribute_form else None

        return super().from_dict(data, connection, to_snake_case=False)


@dataclass
class Expression(Dictable):
    """Class representation of Expression.

    A generic specification for a calculation stored within a metadata object.
    The expression is represented as a tree over nodes. Most internal nodes
    (called operator nodes) are defined by applying a function to the operator's
    child nodes.

    Usually an expression must be non-empty. But in a few cases, most notably
    a filter expression it is valid for an expression to contain no nodes
    at all. An expression is empty if and only if it does not have a tree
    property.

    Alternatively the client may prefer to handle an expression as a list of
    tokens. Each token represents part of the raw text of the expression,
    in some cases annotated with additional information.

    Attributes:
        text: Read only. Human-readable description of the expression.
            It is generated from the current specification of the expression.

        tokens: Optional list, used if the expression is to be presented as
            a stream of tokens.

        tree: Representation of an expression as a tree of nodes, instance of
            one of classes inheriting from ExpressionNode class
    """

    _FROM_DICT_MAP = {
        'tokens': [Token],
        'tree': ExpressionNode.dispatch,
    }

    text: str | None = None
    tokens: list[Token] | None = None
    tree: ExpressionNode | None = None

    def __repr__(self):
        return f"Expression(text='{self.text}')"


def list_functions(
    connection: 'Connection',
    name: str | None = None,
    to_dictionary: bool = False,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project_id: str | None = None,
    project_name: str | None = None,
    limit: int | None = None,
    **filters,
) -> list['SchemaObjectReference'] | list[dict]:
    """Get list of SchemaObjectReference objects or dicts representing
    functions. Optionally filter functions by specifying 'name'.

    When attribute is: `function` for Operator node is set to Function.CUSTOM or
    Function.THIRD_PARTY, `custom_function` attribute is required. This function
    give list of SchemaObjectReference object that can be used as values for
    `custom_function`.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name_begins = ?onny will return Sonny and Tonny

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of functions being searched
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns SchemaObjectReference objects
        search_pattern (SearchPattern enum or int, optional): pattern to
            search for, such as Begin With or Exactly. Possible values are
            available in ENUM mstrio.object_management.SearchPattern.
            Default value is CONTAINS (4).
        project_id (string, optional): Project ID
        project_name(string, optional): Project name
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        **filters: Available filter parameters:
            id str: Attribute's id
            name str: Attribute's name
            date_created str: format: 2001-01-02T20:48:05.000+0000
            date_modified str: format: 2001-01-02T20:48:05.000+0000
            version str: function's version
            owner dict: e.g. {'id': <user's id>, 'name': <user's name>},
                with one or both of the keys: id, name
            acg str | int: access control group

    Returns:
        list with SchemaObjectReference objects or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    objects = search_operations.full_search(
        connection,
        object_types=ObjectTypes.FUNCTION,
        project=project_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )

    if to_dictionary:
        return objects

    objects = [
        {
            **obj,
            'object_id': obj.get('id'),
            'sub_type': (
                ObjectSubType.FUNCTION
                if obj.get('subtype') == ObjectSubTypes.FUNCTION.value
                else None
            ),
        }
        for obj in objects
    ]

    return [SchemaObjectReference.from_dict(obj, connection) for obj in objects]



from typing import TYPE_CHECKING

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.version_helper import is_server_min_version

if TYPE_CHECKING:
    from mstrio.utils.sessions import FuturesSessionWithRenewal

CUBE_FIELDS = '-data.metricValues.extras,-data.metricValues.formatted'


@ErrorHandler(err_msg="Error getting report {report_id} definition. Check report ID.")
def report_definition(connection, report_id):
    """Get the definition of a specific report, including attributes and
    metrics. This in-memory report definition provides information about all
    available objects without actually running any data query/report. The
    results can be used by other requests to help filter large datasets and
    retrieve values dynamically, helping with performance and scalability.

    Args:
        connection: MicroStrategy REST API connection object
        report_id (str): Unique ID of the report

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    return connection.get(endpoint=f'/api/v2/reports/{report_id}')


@ErrorHandler(err_msg="Error getting report {report_id} contents.")
def report_instance(connection, report_id, body=None, offset=0, limit=5000):
    """Get the results of a newly created report instance. This in-memory
    report instance can be used by other requests.

    Args:
        connection: MicroStrategy REST API connection object.
        report_id (str): Unique ID of the report you wish to extract information
            from.
        offset (int, optional): Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    if body is None:
        body = {}
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    return connection.post(
        endpoint=f'/api/v2/reports/{report_id}/instances/', json=body, params=params
    )


@ErrorHandler(err_msg="Error getting cube contents.")
def report_instance_id(connection, report_id, instance_id, offset=0, limit=5000):
    """Get the results of a previously created report instance, using the in-
    memory report instance created by a POST /api/reports/{reportId}/instances
    request.

    Args:
        connection: MicroStrategy REST API connection object
        report_id (str): Unique ID of the report you wish to extract information
            from.
        instance_id (str): Unique ID of the in-memory instance of a published
            report.
        offset (int): Optional. Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    return connection.get(
        endpoint=f'/api/v2/reports/{report_id}/instances/{instance_id}', params=params
    )


def report_instance_id_coroutine(
    future_session: "FuturesSessionWithRenewal",
    report_id,
    instance_id,
    offset=0,
    limit=5000,
):
    """Get the future of a previously created instance for a specific report
    asynchronously, using the in-memory instance created by report_instance().

    Returns:
        Complete Future object.
    """
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(future_session.connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    endpoint = f'/api/v2/reports/{report_id}/instances/{instance_id}'
    future = future_session.get(endpoint=endpoint, params=params)
    return future


@ErrorHandler(err_msg="Error retrieving attribute {attribute_id} elements.")
def report_single_attribute_elements(
    connection, report_id, attribute_id, offset=0, limit=200000
):
    """Get elements of a specific attribute of a specific report.

    Args:
        connection: MicroStrategy REST API connection object.
        report_id (str): Unique ID of the report you wish to extract information
            from.
        attribute_id (str): Unique ID of the attribute in the report.
        offset (int): Optional. Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.
    Returns:
        Complete HTTP response object
    """
    endpoint = f'/api/reports/{report_id}/attributes/{attribute_id}/elements'
    return connection.get(endpoint=endpoint, params={'offset': offset, 'limit': limit})


def report_single_attribute_elements_coroutine(
    future_session: "FuturesSessionWithRenewal",
    report_id,
    attribute_id,
    offset=0,
    limit=200000,
):
    """Get elements of a specific attribute of a specific report.

    Args:
        future_session(object): Future Session object to call MicroStrategy REST
            Server asynchronously
        report_id (str): Unique ID of the report you wish to extract information
            from.
        attribute_id (str): Unique ID of the attribute in the report.
        offset (int): Optional. Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.
    Returns:
        Complete Future object
    """
    endpoint = f'/api/reports/{report_id}/attributes/{attribute_id}/elements'
    future = future_session.get(
        endpoint=endpoint, params={'offset': offset, 'limit': limit}
    )
    return future


@ErrorHandler(err_msg="Error getting collection of prompts for report {report_id}")
def get_report_prompts(connection, report_id, closed=None, fields=None):
    """Get the collection of prompts and their respective definitions from a
    report.

    Args:
        connection: MicroStrategy REST API connection object.
        report_id (str): Unique ID of the report you wish
            to extract information from.
        closed: Prompt status, true means get closed prompt,
            false means get open prompt
        fields: Comma-separated, top-level field whitelist
            that allows the client to selectively retrieve
            part of the response model.

    """
    endpoint = f'/api/reports/{report_id}/prompts'
    return connection.get(
        endpoint=endpoint, params={'closed': closed, 'fields': fields}
    )


@ErrorHandler(err_msg="Error getting prompted report {report_id} instance.")
def get_prompted_instance(connection, report_id, instance_id, closed=None, fields=None):
    """Get the collection of prompts and their respective definitions from a
    report instance. This endpoint will return data only when the report
    instance has prompt which need to be answered.

    Args:
        connection: MicroStrategy REST API connection object.
        report_id (str): Unique ID of the report you wish
            to extract information from.
        instance_id (str): Unique ID of the in-memory instance of a published
            report.
        closed(bool): Prompt status, true means get closed prompt,
            false means get open prompt
        fields: Comma-separated, top-level field whitelist
            that allows the client to selectively retrieve
            part of the response model.

    """
    endpoint = f'/api/reports/{report_id}/instances/{instance_id}/prompts'
    return connection.get(
        endpoint=endpoint, params={'closed': closed, 'fields': fields}
    )


@ErrorHandler(
    err_msg="Error getting prompted report {report_id} attribute element prompt."
)
def get_report_attribute_element_prompt(
    connection: 'Connection',
    report_id: str,
    instance_id: str,
    prompt_id: str,
    offset: int = 0,
    limit: int = 100,
    search_pattern: str | None = None,
    fields: str | None = None,
):
    """Get available attribute element for attribute element prompt"""
    endpoint = (
        f'/api/reports/{report_id}/instances/{instance_id}/prompts/{prompt_id}/elements'
    )
    return connection.get(
        endpoint=endpoint,
        params={
            'offset': offset,
            'limit': limit,
            'searchPattern': search_pattern,
            'fields': fields,
        },
    )



from typing import TYPE_CHECKING

from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.sessions import FuturesSessionWithRenewal

if TYPE_CHECKING:
    from mstrio.connection import Connection


@unpack_information
@ErrorHandler("Error getting the table with ID: {id}")
def get_table(
    connection: 'Connection',
    id: str,
    project_id: str | None = None,
    changeset_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Get a detailed definition for a specified table.

    Args:
        connection (object): MicroStrategy REST API connection object
        id (str): Table ID
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        endpoint=f'/api/model/tables/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'X-MSTR-Changeset': changeset_id},
        params={'fields': fields},
    )


@unpack_information
@ErrorHandler("Error listing tables")
def get_tables(
    connection: 'Connection',
    project_id: str | None = None,
    changeset_id: str | None = None,
    limit: int = None,
    offset: int = 0,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Get a list of all tables.

    Args:
        connection (object): MicroStrategy REST API connection object
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use None for no limit.
            Default is None.
        offset (int, optional): Starting point within the collection of
            returned results. Used to control paging behavior. Default is 0.
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        endpoint='/api/model/tables',
        headers={'X-MSTR-ProjectID': project_id, 'X-MSTR-Changeset': changeset_id},
        params={'limit': limit, 'offset': offset, 'fields': fields},
    )


@unpack_information
@ErrorHandler("Error updating the table with ID: {id}")
def patch_table(
    connection: 'Connection',
    id: str,
    body: dict | None = None,
    column_merge_option: str | None = None,
    fields: dict | None = None,
    error_msg: str | None = None,
):
    """Update a detailed definition for a specified table in the changeset

    Args:
        connection (object): MicroStrategy REST API connection object
        id (str): Table ID
        body (dict): Table update body
        column_merge_option (str, optional): Defines a column merge option
            Available values: 'reuse_any', 'reuse_compatible_data_type',
            'reuse_matched_data_type'.
            If 'reuse_any', updates the column body type to use the most
                recent column definition.
            If 'reuse_compatible_data_type', updates the column body type
                to use the body type with the largest precision or scale.
            If 'reuse_matched_data_type', renames the column in newly added
                table to allow it to have different body types.
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.patch(
            endpoint=f'/api/model/tables/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
            params={'columnMergeOption': column_merge_option, 'fields': fields},
        )


@unpack_information
@ErrorHandler("Error creating the table")
def post_table(
    connection: 'Connection',
    data: dict,
    check_secondary_data_source_table: bool | None = None,
    column_merge_option: str | None = None,
    table_prefix_option: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Update a detailed definition for a specified table in the changeset

    Args:
        connection (object): MicroStrategy REST API connection object
        data (dict): Table creation data
        check_secondary_data_source_table (bool, optional):
            Available values: 'true', 'false'
            If 'true', finds compatible tables in the project.
                If a compatible table is found, the compatible table object
                information is returned. If no table is found,
                a new table is created.
            If 'false', a new table is created.
        column_merge_option (str, optional): Defines a column merge option
            Available values: 'reuse_any', 'reuse_compatible_data_type',
            'reuse_matched_data_type'.
            If 'reuse_any', updates the column data type to use the most
                recent column definition.
            If 'reuse_compatible_data_type', updates the column data type
                to use the data type with the largest precision or scale.
            If 'reuse_matched_data_type', renames the column in newly added
                table to allow it to have different data types.
        table_prefix_option (str, optional): Define the table prefix
            Available values: 'none', 'add_default_prefix', 'add_namespace'.
            If 'none', do not set table prefix.
            If 'add_default_prefix', applies the default prefix setting on
                warehouse catalog.
            If 'add_namespace', create a prefix same with namespace.
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint='/api/model/tables',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=data,
            params={
                'checkSecondaryDataSourceTable': (
                    'true' if check_secondary_data_source_table else 'false'
                ),
                'columnMergeOption': column_merge_option,
                'tablePrefixOption': table_prefix_option,
                'fields': fields,
            },
        )


@ErrorHandler("Error fetching available Warehouse Tables")
def get_available_warehouse_tables(
    connection: 'Connection',
    datasource_id: str,
    namespace_id: str,
    project_id: str | None = None,
    refresh: bool | None = None,
    error_msg: str | None = None,
):
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        endpoint=(
            f'/api/datasources/{datasource_id}/catalog/namespaces/{namespace_id}/tables'
        ),
        headers={'X-MSTR-ProjectID': project_id},
        params={'refresh': refresh},
    )


def get_table_async(
    future_session: FuturesSessionWithRenewal,
    id: str,
    changeset_id: str | None = None,
    project_id: str = None,
    fields=None,
):
    """Get definition of a single table by id, using FuturesSessionWithRenewal

    Args:
        future_session(object): `FuturesSessionWithRenewal` object to call
            MicroStrategy REST Server asynchronously
        id: ID of a table
        changeset_id: ID of a changeset
        project_id: Id of a project
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.
    Return:
        Complete Future object.
    """
    return future_session.get(
        endpoint=f'/api/model/tables/{id}',
        headers={'X-MSTR-MS-Changeset': changeset_id, 'X-MSTR-ProjectID': project_id},
        params={'fields': fields},
    )


def get_available_warehouse_tables_async(
    future_session: FuturesSessionWithRenewal, datasource_id: str, namespace_id: str
):
    return future_session.get(
        endpoint=f'/api/datasources/{datasource_id}/catalog/namespaces/'
        f'{namespace_id}/tables',
        headers={'X-MSTR-ProjectID': future_session.connection.project_id},
    )


@ErrorHandler(err_msg="Error updating the structure of the table with ID: {id}")
def update_structure(
    connection: 'Connection',
    id: str,
    column_merge_option: str | None = None,
    fields: str | None = None,
    ignore_table_prefix: bool | None = None,
):
    """Update physical table structure.
    If table is missing, table cannot be updated. It will do nothing
    if the physical table is a free form sql table. If the table has
    secondary datasource, each physical table's schema in the secondary
    datasource list will also be checked. The update will be rejected
    if the new table structure does not meet secondary datasource condition
    (column missing, data type incompatible).

    Args:
        connection (object): MicroStrategy REST API connection object
        id (str): Logical table's ID. An identifier for the logical table
            object that the client wishes to invoke. The model service does
            not distinguish between logical and physical tables. A physical
            table is accessed as part of a logical table based on the physical
            table field.
        column_merge_option (str, optional): Defines a column merge option
            Available values: 'reuse_any', 'reuse_compatible_data_type',
            'reuse_matched_data_type'.
            If 'reuse_any', updates the column body type to use the most
                recent column definition.
            If 'reuse_compatible_data_type', updates the column body type
                to use the body type with the largest precision or scale.
            If 'reuse_matched_data_type', renames the column in newly added
                table to allow it to have different body types.
        fields (list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        ignore_table_prefix (bool, optional): If true, get all tables under
            current DB. There are three following situations:

                - If there is only one table that has same name as updated
                    table, update table structure using this table.
                - If there is no table that has same name as updated table,
                    throw error.
                - If there are multiple tables has same name as updated table,
                    throw error.

            If false, remain the current behavior.

            If not set, get the setting value from warehouse catalog. This
                behavior is same as column merge options.
    Returns:
        Complete HTTP response object.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint=f'/api/model/tables/{id}/physicalTable/refresh',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={
                'columnMergeOption': column_merge_option,
                'fields': fields,
                'ignoreTablePrefix': (
                    str(ignore_table_prefix).lower()
                    if isinstance(ignore_table_prefix, bool)
                    else ignore_table_prefix
                ),
            },
        )



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.utils.sessions import FuturesSessionWithRenewal


@ErrorHandler(err_msg="Error listing Contact Groups.")
def get_contact_groups(
    connection: 'Connection',
    offset: int = 0,
    limit: int = 1000,
    error_msg: str | None = None,
):
    """Get a list of all contact groups that the authenticated user has access
        to.

    Args:
        connection: MicroStrategy REST API connection object
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 1000.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = '/api/contactGroups'
    params = {'offset': offset, 'limit': limit}
    return connection.get(endpoint=endpoint, params=params)


def get_contact_groups_async(
    future_session: 'FuturesSessionWithRenewal',
    offset: int = 0,
    limit: int = 1000,
):
    """Get a list of all contact groups that the authenticated user has access
        to.

    Args:
        future_session(object): Future Session object to call MicroStrategy REST
            Server asynchronously
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 1000.

    Returns:
        Complete Future object.
    """
    endpoint = '/api/contactGroups'
    params = {'offset': offset, 'limit': limit}
    future = future_session.get(endpoint=endpoint, params=params)
    return future


@ErrorHandler(err_msg="Error creating Contact Group.")
def create_contact_group(
    connection: 'Connection', body: dict, error_msg: str | None = None
):
    """Create a new contact group.

    Args:
        connection: MicroStrategy REST API connection object
        body: Contact group creation body
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    endpoint = '/api/contactGroups'
    return connection.post(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error getting Contact Group with ID {id}")
def get_contact_group(connection: 'Connection', id: str, error_msg: str | None = None):
    """Get contact group by a specific id.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact group
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = f'/api/contactGroups/{id}'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Error updating Contact Group with ID {id}")
def update_contact_group(
    connection: 'Connection', id: str, body: dict, error_msg: str | None = None
):
    """Update a contact group.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact group
        body: Contact group update info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = f'/api/contactGroups/{id}'
    return connection.put(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error deleting Contact Group with ID {id}")
def delete_contact_group(
    connection: 'Connection', id: str, error_msg: str | None = None
):
    """Delete a contact group.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact group
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    endpoint = f'/api/contactGroups/{id}'
    return connection.delete(endpoint=endpoint)



def _map_data_type(datatype):
    if datatype == 'object':
        return "STRING"
    elif datatype in ['int64', 'int32']:
        return "INTEGER"
    elif datatype in ['float64', 'float32']:
        return "DOUBLE"
    elif datatype == 'bool':
        return "BOOL"
    elif datatype == 'datetime64[ns]':
        return 'DATETIME'


def formjson(df, table_name, as_metrics=None, as_attributes=None):
    def _form_column_headers(_col_names, _col_types):
        return [{'name': n, 'dataType': t} for n, t in zip(_col_names, _col_types)]

    def _form_attribute_list(_attributes):
        return [
            {
                'name': n,
                'attributeForms': [
                    {
                        'category': 'ID',
                        'expressions': [{'formula': table_name + "." + n}],
                    }
                ],
            }
            for n in _attributes
        ]

    def _form_metric_list(_metrics):
        return [
            {
                'name': n,
                'dataType': 'number',
                'expressions': [{'formula': table_name + "." + n}],
            }
            for n in _metrics
        ]

    col_names = list(df.columns)
    col_types = list(map(_map_data_type, list(df.dtypes.values)))

    # Adjust attributes/metrics mapping if new mappings were provided in
    # as_metrics and as_attributes
    attributes = []
    metrics = []
    for _name, _type in zip(col_names, col_types):
        if _type in ['DOUBLE', 'INTEGER']:  # metric
            if as_attributes is not None and _name in as_attributes:
                attributes.append(_name)
            else:
                metrics.append(_name)
        else:  # attribute
            if as_metrics is not None and _name in as_metrics:
                metrics.append(_name)
            else:
                attributes.append(_name)

    column_headers = _form_column_headers(_col_names=col_names, _col_types=col_types)
    attribute_list = _form_attribute_list(_attributes=attributes)
    metric_list = _form_metric_list(_metrics=metrics)

    return column_headers, attribute_list, metric_list



from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error creating Contact.")
def create_contact(connection: Connection, body: dict, error_msg: str | None = None):
    """Create a new contact.

    Args:
        connection: MicroStrategy REST API connection object
        body: Contact creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    endpoint = '/api/contacts/'
    return connection.post(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error getting Contact with ID {id}")
def get_contact(connection: Connection, id: str, error_msg: str | None = None):
    """Get contact by a specific id.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = f'/api/contacts/{id}'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Error deleting Contact with ID {id}")
def delete_contact(connection: Connection, id: str, error_msg: str | None = None):
    """Delete a contact.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    endpoint = f'/api/contacts/{id}'
    return connection.delete(endpoint=endpoint)


@ErrorHandler(err_msg="Error updating Contact with ID {id}")
def update_contact(
    connection: Connection, id: str, body: dict, error_msg: str | None = None
):
    """Update a contact.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact
        body: Contact update info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = f'/api/contacts/{id}'
    return connection.put(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error getting Contacts.")
def get_contacts(
    connection: Connection,
    offset: int = 0,
    limit: int = -1,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Get information for all contacts.

    Args:
        connection: MicroStrategy REST API connection object
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = '/api/contacts/'
    params = {'fields': fields, 'offset': offset, 'limit': limit}
    return connection.get(endpoint=endpoint, params=params)



"""This is the demo script to show how administrator can manage server and the
cluster with its nodes.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.server import Cluster, Environment, Project

from mstrio.connection import get_connection
from mstrio.utils.wip import module_wip, WipLevels

# For some methods to work, a connection to environment
# using user's credentials is needed.
module_wip(globals(), level=WipLevels.WARNING)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# get the cluster for given connection
clstr = Cluster(connection=conn)

# save topologies of services or nodes from this cluster into a dataframe
nodes_topology_df = clstr.nodes_topology()
services_topology_df = clstr.services_topology()

# get list of services grouped by nodes or by services
services_by_nodes = clstr.list_services(group_by='nodes')
services_by_services = clstr.list_services(group_by='services')

# Define variables which can be used in a script
NODE_NAME = $node_name
NODE_NAME_2 = $node_name_2

# get list of nodes (information about projects within each node is given there)
nodes = clstr.list_nodes(to_dictionary=True)
# remove/add a node from/to a cluster (node with the given name should exist)
clstr.remove_node(node=NODE_NAME)
clstr.add_node(node=NODE_NAME)

# get name of default (primary) node of the cluster and set new default node
clstr.default_node
clstr.set_primary_node(node=NODE_NAME)

# update node settings or reset them to default values
clstr.update_node_settings(
    node=NODE_NAME, load_balance_factor=99, initial_pool_size=511, max_pool_size=1023
)
clstr.reset_node_settings(node=NODE_NAME)

# Define a variable which can be later used in a script
SERVICE_NAME = $service_name

# stop/start service on selected nodes (error will be thrown in case of wrong
# names of service or nodes)
clstr.stop(service=SERVICE_NAME, nodes=[NODE_NAME, NODE_NAME_2])
clstr.start(service=SERVICE_NAME, nodes=[NODE_NAME, NODE_NAME_2])

# list all projects available for the given connection (it is possible via
# class Cluster or Environment)
env = Environment(connection=conn)
projects = env.list_projects()
projects = clstr.list_projects()

# load or unload chosen project (it is possible via class Cluster or
# Project)
project = Project(connection=conn, name=PROJECT_NAME)
project.load()
project.unload()

# via Cluster can we also specify on which node(s) project will be loaded
# or unloaded
clstr.load_project(project=PROJECT_NAME, on_nodes=[NODE_NAME, NODE_NAME_2])
clstr.unload_project(project=PROJECT_NAME, on_nodes=[NODE_NAME, NODE_NAME_2])

# get settings of a server as a dataframe
server_settings_df = env.server_settings.to_dataframe

# Define a variable which can be later used in a script
FILE_NAME = $file_name  # file name with extension 'csv', 'json' or 'pickle'

# save/load settings of a server to/from a file (format can be 'csv', 'json' or
# 'pickle')
env.server_settings.to_csv(name=FILE_NAME)
env.server_settings.import_from(file=FILE_NAME)

# update some settings of a server
env.server_settings.allowUserLoginWithFullName = True
env.server_settings.minCharsPasswordChanges = 1
env.server_settings.update()



from enum import IntFlag
from typing import Any, TypeVar

from mstrio.api import browsing
from mstrio.connection import Connection
from mstrio.utils.entity import CopyMixin, Entity, MoveMixin, ObjectTypes
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import deprecation_warning, fetch_objects, get_valid_project_id
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin


class ShortcutInfoFlags(IntFlag):
    DssDossierShortcutInfoBookmark = 0b10
    DssDossierShortcutInfoTOC = 0b01
    DssDossierShortcutInfoDefault = 0b00
    DssDashboardShortcutInfoBookmark = 0b10
    DssDashboardShortcutInfoTOC = 0b01
    DssDashboardShortcutInfoDefault = 0b00


class Shortcut(Entity, CopyMixin, MoveMixin, TranslationMixin):
    _OBJECT_TYPE = ObjectTypes.SHORTCUT_TYPE
    _API_GETTERS = {
        (
            'name',
            'id',
            'project_id',
            'owned_by_current_user',
            'target',
            'encode_html_content',
            'current_page_key',
            'shared_time',
            'last_viewed_time',
            'last_modified_time',
            'stid',
            'current_bookmark',
            'prompted',
            'datasets_cache_info_hash',
            'shortcut_info_flag',
            'dossier_version_hash',
        ): browsing.get_shortcut,
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
            'hidden',
        ): objects_processors.get_info,
    }
    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'shortcut_info_flag': ShortcutInfoFlags}
    _API_PATCH: dict = {
        ('name', 'description', 'abbreviation', 'hidden', 'folder_id'): (
            objects_processors.update,
            'partial_put',
        )
    }

    def __init__(
        self,
        connection: Connection,
        id: str,
        project_id: str = None,
        project_name: str = None,
        shortcut_info_flag: (
            ShortcutInfoFlags | int
        ) = ShortcutInfoFlags.DssDossierShortcutInfoTOC,
    ):
        """Initialize the Shortcut object and populate it with I-Server data.

        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Note:
            When `project_id` is `None` and `project_name` is `None`, then
            its value is overwritten by `project_id` from `connection` object.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Shortcut ID
            project_id: ID of the project that the shortcut is in
            project_name: Project name
            shortcut_info_flag: flag indicating what information about shortcut
                should be loaded
        """
        if id is None:
            raise AttributeError("Please specify 'id' parameter in the constructor.")
        else:
            project_id = get_valid_project_id(
                connection=connection,
                project_id=project_id,
                project_name=project_name,
                with_fallback=not project_name,
            )
            super().__init__(
                connection=connection,
                object_id=id,
                project_id=project_id,
                shortcut_info_flag=get_enum_val(shortcut_info_flag, ShortcutInfoFlags),
            )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._owned_by_current_user = kwargs.get('owned_by_current_user')
        self._target = kwargs.get('target')
        self._encode_html_content = kwargs.get('encode_html_content')
        self._shortcut_info_flag = (
            ShortcutInfoFlags(kwargs["shortcut_info_flag"])
            if kwargs.get("shortcut_info_flag")
            else None
        )
        self._current_page_key = kwargs.get('current_page_key')
        self._shared_time = kwargs.get('shared_time')
        self._last_viewed_time = kwargs.get('last_viewed_time')
        self._last_modified_time = kwargs.get('last_modified_time')
        self._stid = kwargs.get('stid')
        self._current_bookmark = kwargs.get('current_bookmark')
        self._prompted = kwargs.get('prompted')
        self._datasets_cache_info_hash = kwargs.get('datasets_cache_info_hash')
        self._dossier_version_hash = kwargs.get('dossier_version_hash')
        self._dashboard_version_hash = kwargs.get('dossier_version_hash')

    T = TypeVar('T')

    @classmethod
    def from_dict(cls: T, source: dict[str, Any], connection: Connection) -> T:
        """Instantiate a Shortcut from a dict source."""
        return super(Entity, cls).from_dict(source=source, connection=connection)

    @property
    def owned_by_current_user(self):
        return self._owned_by_current_user

    @property
    def target(self):
        return self._target

    @property
    def encode_html_content(self):
        return self._encode_html_content

    @property
    def shortcut_info_flag(self):
        return self._shortcut_info_flag

    @property
    def current_page_key(self):
        return self._current_page_key

    @property
    def shared_time(self):
        return self._shared_time

    @property
    def last_viewed_time(self):
        return self._last_viewed_time

    @property
    def last_modified_time(self):
        return self._last_modified_time

    @property
    def stid(self):
        return self._stid

    @property
    def current_bookmark(self):
        return self._current_bookmark

    @property
    def prompted(self):
        return self._prompted

    @property
    def datasets_cache_info_hash(self):
        return self._datasets_cache_info_hash

    @property
    def dossier_version_hash(self):
        deprecation_warning(
            'property `dossier_version_hash`',
            'property `dashboard_version_hash`',
            '11.5.03',
            False,
        )
        return self._dossier_version_hash

    @property
    def dashboard_version_hash(self):
        return self._dashboard_version_hash


def get_shortcuts(
    connection: Connection,
    shortcut_ids: list[str],
    project_id: str = None,
    project_name: str = None,
    shortcut_info_flag: (
        ShortcutInfoFlags | int
    ) = ShortcutInfoFlags.DssDossierShortcutInfoDefault,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list[dict] | list[Shortcut]:
    """Retrieve information about specific published shortcuts
     in specific project.

     Specify either `project_id` or `project_name`.
     When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
         When `project_id` is `None` and `project_name` is `None`,
         then its value is overwritten by `project_id` from `connection`
         object.

     Args:
         shortcut_ids: ids of target shortcuts
         project_id: id of project that the shortcuts are in
         project_name: Project name
         shortcut_info_flag: a single ShortcutInfoFlags that describes what
           exact info are to be fetched
         to_dictionary: parameter describing output format
         limit (int): limit the number of elements returned.
             If `None` (default), all objects are returned.
     Return:
         list of dictionaries or Shortcut objects,
           depending on `to_dictionary` parameter
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    shortcuts = fetch_objects(
        connection=connection,
        api=browsing.get_shortcuts,
        dict_unpack_value="shortcuts",
        limit=limit,
        filters=filters,
        body=[{"projectId": project_id, "shortcutIds": shortcut_ids}],
        shortcut_info_flag=get_enum_val(shortcut_info_flag, ShortcutInfoFlags),
    )

    if to_dictionary:
        return shortcuts
    else:
        return [
            Shortcut.from_dict(source=short, connection=connection)
            for short in shortcuts
        ]



import logging
import time
from enum import Enum, IntEnum
from typing import TYPE_CHECKING, Optional

from pandas import DataFrame, Series
from tqdm import tqdm

from mstrio import config
from mstrio.api import datasources as datasources_api
from mstrio.api import monitors, projects
from mstrio.connection import Connection
from mstrio.helpers import IServerError
from mstrio.server.project_languages import (
    DataLanguageSettings,
    MetadataLanguageSettings,
)
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.response_processors import datasources as datasources_processors
from mstrio.utils.response_processors import projects as projects_processors
from mstrio.utils.settings.base_settings import BaseSettings
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import is_server_min_version, method_version_handler
from mstrio.utils.vldb_mixin import ModelVldbMixin
from mstrio.utils.wip import wip

if TYPE_CHECKING:
    from mstrio.datasources import DatasourceInstance

logger = logging.getLogger(__name__)


class ProjectStatus(IntEnum):
    ACTIVE = 0
    ERRORSTATE = -3
    EXECIDLE = 1
    FULLIDLE = 7
    OFFLINE = -1
    OFFLINEPENDING = -2
    ONLINEPENDING = -4
    REQUESTIDLE = 2
    WHEXECIDLE = 4


class IdleMode(Enum):
    """Used to specify the exact behaviour of `idle` request.

    `REQUEST` (Request Idle): all executing and queued jobs finish executing
        and any newly submitted jobs are rejected.
    `EXECUTION` (Execution Idle for All Jobs): all executing, queued, and
        newly submitted jobs are placed in the queue, to be executed when
        the project resumes.
    `WAREHOUSEEXEC` (Execution Idle for Warehouse jobs): all executing,
        queued, and newly submitted jobs that require SQL to be submitted
        to the data warehouse are placed in the queue, to be executed when
        the project resumes. Any jobs that do not require SQL to be
        executed against the data warehouse are executed.
    `FULL` (Request Idle and Execution Idle for All jobs): all executing and
        queued jobs are canceled, and any newly submitted jobs are rejected.
    `PARTIAL` (Request Idle and Execution Idle for Warehouse jobs): all
        executing and queued jobs that do not submit SQL against the data
        warehouse are canceled, and any newly submitted jobs are rejected.
        Any currently executing and queued jobs that do not require SQL to
        be executed against the data warehouse are executed.
    """

    REQUEST = "request_idle"
    EXECUTION = "exec_idle"
    WAREHOUSEEXEC = "wh_exec_idle"
    FULL = "partial_idle"
    PARTIAL = "full_idle"


def compare_project_settings(
    projects: list["Project"], show_diff_only: bool = False
) -> DataFrame:
    """Compares settings of project objects.

    Args:
        projects (List[Project]): List of project objects
            to compare.
        show_diff_only (bool, optional): Whether to display all settings
            or only different from first project in list.
    """
    to_be_compared = {}
    df = DataFrame({}, columns=['initial'])

    for project in projects:
        to_be_compared[project.name] = project.settings.to_dataframe().reset_index()
        if df.empty:
            df.columns = ['setting']
            df['setting'] = to_be_compared[project.name]['setting']
        df[project.name] = to_be_compared[project.name]['value']

    if show_diff_only:
        index = Series([True] * len(df['setting']))
        base = projects[0].name
        for project_name in to_be_compared.keys():
            compare = df[base].eq(df[project_name])
            index = compare & index
        df = df[~index]
        if df.empty and config.verbose:
            project_names = list(to_be_compared.keys())
            project_names.remove(base)
            msg = (
                f"There is no difference in settings between project '{base}' and "
                f"remaining projects: '{project_names}'"
            )
            logger.info(msg)
    return df


class Project(Entity, ModelVldbMixin, DeleteMixin, TranslationMixin):
    """Object representation of MicroStrategy Project (Project) object.

    Attributes:
        connection: A MicroStrategy connection object
        settings: Project settings object
        id: Project ID
        name: Project name
        description: Project description
        alias: Project alias
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        nodes: List of nodes on which project is loaded
        date_created: Creation time, "yyyy-MM-dd HH:mm:ss" in UTC
        date_modified: Last modification time, "yyyy-MM-dd
        version: Version ID
        owner: owner ID and name
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
        status: Project
        ancestors: List of ancestor folders
        data_language_settings: Project data language settings
        metadata_language_settings: Project metadata language settings
    """

    _OBJECT_TYPE = ObjectTypes.PROJECT
    _API_GETTERS = {
        **Entity._API_GETTERS,
        ('status', 'alias'): projects.get_project,
        'nodes': monitors.get_node_info,
        (
            'data_language_settings',
            'metadata_language_settings',
        ): projects_processors.get_project_internalization,
    }
    _API_DELETE = staticmethod(projects.delete_project)
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'status': ProjectStatus,
        'data_language_settings': DataLanguageSettings.from_dict,
        'metadata_language_settings': MetadataLanguageSettings.from_dict,
    }
    _STATUS_PATH = "/status"
    _MODEL_VLDB_API = {
        'GET_ADVANCED': projects.get_vldb_settings,
        'PUT_ADVANCED': projects.update_vldb_settings,
        'GET_APPLICABLE': projects.get_applicable_vldb_settings,
    }
    _REST_ATTR_MAP = {
        'data': 'data_language_settings',
        'metadata': 'metadata_language_settings',
    }

    def __init__(
        self,
        connection: Connection,
        name: str | None = None,
        id: str | None = None,
    ) -> None:
        """Initialize Project object by passing `name` or `id`. When `id` is
        provided (not `None`), `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            name: Project name
            id: Project ID
        """

        # initialize either by ID or Project Name
        if id is None and name is None:
            helper.exception_handler(
                "Please specify either 'name' or 'id' parameter in the constructor.",
                exception_type=ValueError,
            )

        if id is None:
            project_list = Project._list_project_ids(connection, name=name)
            if project_list:
                id = project_list[0]
            else:
                helper.exception_handler(
                    f"There is no project with the given name: '{name}'",
                    exception_type=ValueError,
                )

        try:
            super().__init__(connection=connection, object_id=id, name=name)
        except IServerError as e:
            if not self.is_loaded():
                helper.exception_handler(
                    (
                        "Some projects are either unloaded or idled. Change "
                        "status using the 'load()' or 'resume()' method to use "
                        "all functionality."
                    ),
                    exception_type=UserWarning,
                )
            else:
                raise e

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._status = kwargs.get("status")
        self._alias = kwargs.get("alias")
        self._nodes = kwargs.get("nodes")
        self._data_language_settings = kwargs.get("data_language_settings")
        self._metadata_language_settings = kwargs.get("metadata_language_settings")

    @classmethod
    def _create(
        cls,
        connection: Connection,
        name: str,
        description: str | None = None,
        force: bool = False,
    ) -> Optional["Project"]:
        user_input = 'N'
        if not force:
            user_input = input(
                f"Are you sure you want to create new project '{name}'? [Y/N]: "
            )

        if force or user_input == 'Y':
            # Create new project
            with tqdm(
                desc=f"Please wait while Project '{name}' is being created.",
                bar_format='{desc}',
                leave=False,
                disable=config.verbose,
            ):
                projects.create_project(
                    connection, {"name": name, "description": description}
                )
                http_status, i_server_status = 500, 'ERR001'
                while http_status == 500 and i_server_status == 'ERR001':
                    time.sleep(1)
                    response = projects.get_project(
                        connection, name, whitelist=[('ERR001', 500)]
                    )
                    http_status = response.status_code
                    data = response.json()
                    i_server_status = data.get('code')
            if config.verbose:
                logger.info(f"Project '{name}' successfully created.")
            return cls.from_dict(data, connection=connection)
        else:
            return None

    @classmethod
    @method_version_handler('11.2.0000')
    def _list_projects(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list["Project"] | list[dict]:
        msg = "Error getting information for a set of Projects."
        objects = helper.fetch_objects_async(
            connection,
            monitors.get_projects,
            monitors.get_projects_async,
            dict_unpack_value='projects',
            limit=limit,
            chunk_size=1000,
            error_msg=msg,
            filters=filters,
        )
        if to_dictionary:
            return objects
        else:
            projects = [
                cls.from_dict(source=obj, connection=connection) for obj in objects
            ]
            projects_loaded = Project._list_loaded_projects(
                connection, to_dictionary=True
            )
            projects_loaded_ids = [project['id'] for project in projects_loaded]
            unloaded = [
                project for project in projects if project.id not in projects_loaded_ids
            ]

            if unloaded:
                msg = (
                    f"Projects {[project.name for project in unloaded]} are either "
                    f"unloaded or idled. Change status using the 'load()' or 'resume()'"
                    f" method to use all functionality."
                )
                helper.exception_handler(msg, exception_type=UserWarning)
            return projects

    @classmethod
    def _list_project_ids(
        cls, connection: Connection, limit: int | None = None, **filters
    ) -> list[str]:
        project_dicts = Project._list_projects(
            connection=connection,
            to_dictionary=True,
            limit=limit,
            **dict(filters),
        )
        return [project['id'] for project in project_dicts]

    @classmethod
    def _list_loaded_projects(
        cls, connection: Connection, to_dictionary: bool = False, **filters
    ) -> list["Project"] | list[dict]:
        response = projects.get_projects(connection, whitelist=[('ERR014', 403)])
        list_of_dicts = response.json() if response.ok else []
        list_of_dicts = helper.camel_to_snake(list_of_dicts)  # Convert keys
        raw_project = helper.filter_list_of_dicts(list_of_dicts, **filters)

        if to_dictionary:
            # return list of project names
            return raw_project
        else:
            # return list of Project objects
            return [
                cls.from_dict(source=obj, connection=connection) for obj in raw_project
            ]

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        comments: str | None = None,
    ):
        """Alter project name or/and description.

        Args:
            name: new name of the project.
            description: new description of the project.
            comments: long description of the project.
        """
        properties = helper.filter_params_for_func(
            self.alter, locals(), exclude=['self']
        )

        self._alter_properties(**properties)

    def __change_project_state(
        self, func, on_nodes: str | list[str] | None = None, **mode
    ):
        if type(on_nodes) is list:
            for node in on_nodes:
                func(node, **mode)
        elif type(on_nodes) is str:
            func(on_nodes, **mode)
        elif on_nodes is None:
            for node in self.nodes:
                func(node.get('name'), **mode)  # type: ignore
        else:
            helper.exception_handler(
                "'on_nodes' argument needs to be of type: [list[str], str, NoneType]",
                exception_type=TypeError,
            )

    @method_version_handler('11.2.0000')
    def idle(
        self,
        on_nodes: str | list[str] | None = None,
        mode: IdleMode | str = IdleMode.REQUEST,
    ) -> None:
        """Request to idle a specific cluster node. Idle project with mode
        options.

        Args:
            on_nodes: Name of node, if not passed, project will be idled on
                all of the nodes.
            mode: One of: `IdleMode` values.
        """

        def idle_project(node: str, mode: IdleMode):
            body = {
                "operationList": [
                    {"op": "replace", "path": self._STATUS_PATH, "value": mode.value}
                ]
            }
            response = monitors.update_node_properties(
                self.connection, node, self.id, body
            )
            if response.status_code == 202:
                tmp = helper.filter_list_of_dicts(self.nodes, name=node)
                tmp[0]['projects'] = [response.json()['project']]
                self._nodes = tmp
                if tmp[0]['projects'][0]['status'] != mode.value:
                    self.fetch('nodes')
                if config.verbose:
                    logger.info(
                        f"Project '{self.id}' changed status to '{mode}' on node "
                        f"'{node}'."
                    )

        if not isinstance(mode, IdleMode):
            # Previously `mode` was just a string with possible values
            # corresponding to the member names of the current IdleMode enum.
            # This attempts to convert it to avoid breaking backwards compat.
            if mode in IdleMode.__members__.values():
                mode = IdleMode(mode)
            elif mode in IdleMode.__members__:
                mode = IdleMode[mode]
            else:
                helper.exception_handler(
                    "Unsupported mode, please provide a valid `IdleMode` value.",
                    KeyError,
                )

        self.__change_project_state(func=idle_project, on_nodes=on_nodes, mode=mode)

    @method_version_handler('11.2.0000')
    def resume(self, on_nodes: str | list[str] | None = None) -> None:
        """Request to resume the project on the chosen cluster nodes. If
        nodes are not specified, the project will be loaded on all nodes.

        Args:
            on_nodes: Name of node, if not passed, project will be resumed
                on all of the nodes.
        """

        def resume_project(node):
            body = {
                "operationList": [
                    {"op": "replace", "path": self._STATUS_PATH, "value": "loaded"}
                ]
            }
            response = monitors.update_node_properties(
                self.connection, node, self.id, body
            )
            if response.status_code == 202:
                tmp = helper.filter_list_of_dicts(self.nodes, name=node)
                tmp[0]['projects'] = [response.json()['project']]
                self._nodes = tmp
                if tmp[0]['projects'][0]['status'] != 'loaded':
                    self.fetch('nodes')
                if config.verbose:
                    logger.info(f"Project '{self.id}' resumed on node '{node}'.")

        self.__change_project_state(func=resume_project, on_nodes=on_nodes)

    @method_version_handler('11.2.0000')
    def load(self, on_nodes: str | list[str] | None = None) -> None:
        """Request to load the project onto the chosen cluster nodes. If
        nodes are not specified, the project will be loaded on all nodes.

        Args:
            on_nodes: Name of node, if not passed, project will be loaded
                on all of the nodes.
        """

        def load_project(node):
            body = {
                "operationList": [
                    {"op": "replace", "path": self._STATUS_PATH, "value": "loaded"}
                ]
            }
            response = monitors.update_node_properties(
                self.connection, node, self.id, body
            )
            if response.status_code == 202:
                tmp = helper.filter_list_of_dicts(self.nodes, name=node)
                tmp[0]['projects'] = [response.json()['project']]
                self._nodes = tmp
                if tmp[0]['projects'][0]['status'] != 'loaded':
                    self.fetch('nodes')
                if config.verbose:
                    logger.info(f"Project '{self.id}' loaded on node '{node}'.")

        self.__change_project_state(func=load_project, on_nodes=on_nodes)

    @method_version_handler('11.2.0000')
    def unload(self, on_nodes: str | list[str] | None = None) -> None:
        """Request to unload the project from the chosen cluster nodes. If
        nodes are not specified, the project will be unloaded on all nodes.
        The unload action cannot be performed until all jobs and connections
        for project are completed. Once these processes have finished,
        pending project will be automatically unloaded.

        Args:
            on_nodes: Name of node, if not passed, project will be unloaded
                on all of the nodes.
        """

        def unload_project(node):
            body = {
                "operationList": [
                    {"op": "replace", "path": self._STATUS_PATH, "value": "unloaded"}
                ]
            }
            response = monitors.update_node_properties(
                self.connection, node, self.id, body, whitelist=[('ERR001', 500)]
            )
            if response.status_code == 202:
                tmp = helper.filter_list_of_dicts(self.nodes, name=node)
                tmp[0]['projects'] = [response.json()['project']]
                self._nodes = tmp
                if tmp[0]['projects'][0]['status'] != 'unloaded':
                    self.fetch('nodes')
                if config.verbose:
                    logger.info(f"Project '{self.id}' unloaded on node '{node}'.")
            if response.status_code == 500 and config.verbose:  # handle whitelisted
                logger.warning(
                    f"Project '{self.id}' already unloaded on node '{node}'."
                )

        self.__change_project_state(func=unload_project, on_nodes=on_nodes)

    @method_version_handler('11.3.0800')
    def delete(self: Entity) -> bool:
        """Delete project.

        Returns:
            True if project was deleted, False otherwise.
        """
        self._DELETE_CONFIRM_MSG = (
            f"Are you sure you want to delete project "
            f"'{self.name}' with ID: {self._id}?\n"
            "All objects will be permanently deleted. This cannot be undone.\n"
            "Please type the project name to confirm: "
        )
        self._DELETE_SUCCESS_MSG = (
            f"Project '{self.name}' has been successfully deleted."
        )
        self._DELETE_PROMPT_ANSWER = self.name

        return super().delete(force=False)

    @method_version_handler('11.3.0000')
    def register(self, on_nodes: str | list | None = None) -> None:
        """Register project on nodes.

        A registered project will load on node (server) startup.

        Args:
            on_nodes: Name of node, if not passed, project will be loaded
                on all available nodes on startup.
        """
        if on_nodes is None:
            value = [node['name'] for node in self.nodes]
        else:
            on_nodes = on_nodes if isinstance(on_nodes, list) else [on_nodes]
            value = list(set(self.load_on_startup) | set(on_nodes))
        self._register(on_nodes=value)

    @method_version_handler('11.3.0000')
    def unregister(self, on_nodes: str | list | None = None) -> None:
        """Unregister project on nodes.

        An unregistered project will not load on node (server) startup.

        Args:
            on_nodes (str or list, optional): Name of node, if not passed,
                project will not be loaded on any nodes on startup.
        """
        if on_nodes is None:
            value = []
        else:
            on_nodes = on_nodes if isinstance(on_nodes, list) else [on_nodes]
            value = list(set(self.load_on_startup) - set(on_nodes))
        self._register(on_nodes=value)

    def update_settings(self) -> None:
        """Update the current project settings on the I-Server."""
        self.settings.update(self.id)

    def enable_caching(self) -> None:
        """Enable caching settings for the current project on the I-Server."""
        self.settings.enable_caching()
        self.update_settings()

    def disable_caching(self) -> None:
        """Disable caching settings for the current project on the I-Server."""
        self.settings.disable_caching()
        self.update_settings()

    @method_version_handler('11.3.0800')
    def delete_object_cache(self) -> None:
        """Delete object cache for the current project on the I-Server."""
        monitors.delete_caches(self.connection, self.id, 'object')
        if config.verbose:
            logger.info(
                f"Object cache was successfully deleted for project '{self.id}'"
            )

    @method_version_handler('11.3.0800')
    def delete_element_cache(self) -> None:
        """Delete element cache for the current project on the I-Server."""
        monitors.delete_caches(self.connection, self.id, 'element')
        if config.verbose:
            logger.info(
                f"Element cache was successfully deleted for project '{self.id}'"
            )

    def fetch_settings(self) -> None:
        """Fetch the current project settings from the I-Server."""
        self.settings.fetch(self.id)

    def list_caching_settings(self) -> dict:
        """
        Fetch current project settings connected with caching from I-Server
        """
        return self.settings.list_caching_properties()

    def is_loaded(self) -> bool:
        """Check if the project is loaded on any node (server)."""
        loaded = False
        self.fetch('nodes')
        if not isinstance(self.nodes, list):
            helper.exception_handler(
                "Could not retrieve current project status.",
                exception_type=ConnectionError,
            )
        for node in self.nodes:
            projects = node.get('projects')
            if projects:
                status = projects[0].get('status')
                loaded = status == 'loaded'
                if loaded:
                    break
        return loaded

    def get_data_engine_versions(self) -> dict:
        """Fetch the currently available data engine versions for project."""

        return projects.get_engine_settings(self.connection, self.id).json()['engine'][
            'versions'
        ]

    def update_data_engine_version(self, new_version: int) -> None:
        """Update data engine version for project."""

        self.alter_vldb_settings(names_to_values={'AEVersion': new_version})

    def _register(self, on_nodes: list) -> None:
        path = f"/projects/{self.id}/nodes"
        body = {"operationList": [{"op": "replace", "path": path, "value": on_nodes}]}
        projects.update_projects_on_startup(self.connection, body)
        if config.verbose:
            if on_nodes:
                logger.info(f'Project will load on startup of: {on_nodes}')
            else:
                logger.warning('Project will not load on startup.')

    @method_version_handler('11.3.0100')
    def add_datasource(
        self, datasources: list['DatasourceInstance | str']
    ) -> list['DatasourceInstance']:
        """Add datasources to the project.

        Args:
            datasources (list["DatasourceInstance", str]): List of datasource
                objects or datasource IDs to be added to the project.

        Returns:
            List of datasource objects for the project.
        """
        from mstrio.datasources import DatasourceInstance, list_datasource_instances

        operation_list = []
        valid_datasource_ids = [
            datasource.id for datasource in list_datasource_instances(self.connection)
        ]
        datasource_ids = [
            datasource.id if isinstance(datasource, DatasourceInstance) else datasource
            for datasource in datasources
        ]

        for datasource_id in datasource_ids:
            if datasource_id not in valid_datasource_ids:
                raise ValueError(f"Datasource with ID {datasource_id} doesn't exist.")

            operation_list.append(
                {
                    'op': 'add',
                    'path': '/id',
                    'value': datasource_id,
                }
            )

        response = datasources_processors.update_project_datasources(
            connection=self.connection,
            id=self.id,
            body={'operationList': operation_list},
        )

        if config.verbose:
            logger.info(
                f"Datasources were successfully added to the project {self.id}."
            )

        return [
            DatasourceInstance.from_dict(datasource, self.connection)
            for datasource in response
        ]

    @method_version_handler('11.3.0100')
    def remove_datasource(self, datasources: list['DatasourceInstance | str']) -> None:
        """Remove datasources from the project.

        Args:
            datasources (list["DatasourceInstance", str]): List of datasource
                objects or datasource IDs to be removed from the project.
        """
        from mstrio.datasources import DatasourceInstance, list_datasource_instances

        operation_list = []
        valid_datasource_ids = [
            datasource.id
            for datasource in list_datasource_instances(
                self.connection, project=self.id
            )
        ]
        datasource_ids = [
            datasource.id if isinstance(datasource, DatasourceInstance) else datasource
            for datasource in datasources
        ]

        for datasource_id in datasource_ids:
            if datasource_id not in valid_datasource_ids:
                raise ValueError(
                    f"Datasource with ID {datasource_id} does not exist in the "
                    f"project {self.id}."
                )

            operation_list.append(
                {
                    'op': 'remove',
                    'path': '/id',
                    'value': datasource_id,
                }
            )

        datasources_api.update_project_datasources(
            connection=self.connection,
            id=self.id,
            body={'operationList': operation_list},
        )

        if config.verbose:
            logger.info(
                f"Datasources were successfully removed from the project {self.id}."
            )

    def list_properties(self, excluded_properties: list[str] | None = None) -> dict:
        if not is_server_min_version(self.connection, '11.3.1200'):
            excluded_properties = [
                'data_language_settings',
                'metadata_language_settings',
            ] + (excluded_properties or [])
            return super().list_properties(excluded_properties)

        return super().list_properties(excluded_properties)

    @property
    def load_on_startup(self):
        """View nodes (servers) to load project on startup."""
        response = projects.get_projects_on_startup(self.connection).json()
        return response['projects'][self.id]['nodes']

    @property
    def settings(self) -> "ProjectSettings":
        """`Settings` object storing Project settings.

        Settings can be listed by using `list_properties()` method.
        Settings can be modified directly by setting the values in the
        object.
        """

        if not hasattr(self, "_settings"):
            super(Entity, self).__setattr__(
                "_settings", ProjectSettings(self.connection, self.id)
            )
        return self._settings

    @settings.setter
    def settings(self, settings: "ProjectSettings") -> None:
        super(Entity, self).__setattr__("_settings", settings)

    @property
    def status(self):
        return self._status

    @property
    def alias(self):
        return self._alias

    @property
    def nodes(self):
        return self._nodes

    @property
    @method_version_handler(version='11.3.1200')
    def data_language_settings(self) -> DataLanguageSettings:
        if not hasattr(self._data_language_settings, '_connection'):
            self._data_language_settings._connection = self.connection
        if not hasattr(self._data_language_settings, '_project_id'):
            self._data_language_settings._project_id = self.id
        return self._data_language_settings

    @property
    @method_version_handler(version='11.3.1200')
    def metadata_language_settings(self) -> MetadataLanguageSettings:
        if not hasattr(self._metadata_language_settings, '_connection'):
            self._metadata_language_settings._connection = self.connection
        if not hasattr(self._metadata_language_settings, '_project_id'):
            self._metadata_language_settings._project_id = self.id
        return self._metadata_language_settings


class ProjectSettings(BaseSettings):
    """Object representation of MicroStrategy Project (Project) Settings.

    Used to fetch, view, modify, update, export to file, import from file and
    validate Project settings.

    The object can be optionally initialized with `connection` and
    `project_id`, which will automatically fetch the current settings for
    the specified project. If not specified, settings can be loaded from
    file using `import_from()` method. Object attributes (representing settings)
    can be modified manually. Lastly, the object can be applied to any
    project using the `update()` method.

    Attributes:
        connection: A MicroStrategy connection object
    """

    _TYPE = "allProjectSettings"
    _CONVERSION_MAP = {
        'maxCubeSizeForDownload': 'B',
        'maxDataUploadSize': 'B',
        'maxMstrFileSize': 'B',
        'maxHistoryListInboxMessage': 'B',
        'maxMemoryPerDataFetch': 'B',
        'maxElementCacheMemoryConsumption': 'B',
        'maxObjectCacheMemoryConsumption': 'B',
        'maxRWDCacheMemoryConsumption': 'B',
        'maxReportCacheMemoryConsumption': 'B',
        'maxCubeQuota': 'B',
        'maxCubeMemUsage': 'B',
        'maxSqlGenerationMemoryConsumption': 'B',
        'reportCacheLifeTime': 'hour',
        'maxWarehouseJobExecTime': 'sec',
        'maxReportExecutionTime': 'sec',
        'maxScheduledReportExecutionTime': 'sec',
        'statisticsPurgeTimeout': 'sec',
        'maxPromptWaitingTime': 'sec',
        'maxRAMForReportRWDCacheIndex': '%',
        'cubeIndexGrowthUpperBound': '%',
    }
    _CACHING_SETTINGS_TO_ENABLE = (
        "enableReportServerCaching",
        "enableCachingForPromptedReportDocument",
        "enableCachingForNonPromptedReportDocument",
        "enableXmlCachingForReport",
    )
    _CACHING_SETTINGS_TO_DISABLE = _CACHING_SETTINGS_TO_ENABLE + (
        "recordPromptAnswersForCacheMonitoring",
        "enableDocumentOutputCachingInXml",
        "enableDocumentOutputCachingInHtml",
        "enableDocumentOutputCachingInPdf",
        "enableDocumentOutputCachingInExcel",
    )

    def __init__(self, connection: Connection, project_id: str | None = None):
        """Initialize `ProjectSettings` object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            project_id: Project ID
        """
        super(BaseSettings, self).__setattr__('_connection', connection)
        super(BaseSettings, self).__setattr__('_project_id', project_id)
        self._configure_settings()

        if project_id:
            self.fetch()

    def fetch(self, project_id: str | None = None) -> None:
        """Fetch current project settings from I-Server and update this
        `ProjectSettings` object.

        Args:
            project_id: Project ID
        """
        self._check_params(project_id)
        super().fetch()

    def update(self, project_id: str | None = None) -> None:
        """Update the current project settings on I-Server using this
        Settings object.

        Args:
            project_id: Project ID
        """
        self._check_params(project_id)
        set_dict = self._prepare_settings_push()
        response = projects.update_project_settings(
            self._connection, self._project_id, set_dict
        )
        if config.verbose:
            if response.status_code == 200:
                logger.info('Project settings updated.')
            elif response.status_code == 207:
                partial_succ = response.json()
                logging.info(f'Project settings partially successful.\n{partial_succ}')

    def __setattr__(self, key, value) -> None:
        if isinstance(value, Enum):
            value = get_enum_val(value, type(value))

        if isinstance(value, list):
            value = [
                get_enum_val(v, type(v)) if isinstance(v, Enum) else v for v in value
            ]

        super().__setattr__(key, value)

    @wip()
    def enable_caching(self) -> None:
        """
        Enable caching settings for the current project on I-Server
        using this Settings object
        """
        for setting in self._CACHING_SETTINGS_TO_ENABLE:
            if hasattr(self, setting):
                self.__setattr__(setting, True)

    @wip()
    def disable_caching(self) -> None:
        """
        Disable caching settings for the current project on I-Server
        using this Settings object
        """
        for setting in self._CACHING_SETTINGS_TO_DISABLE:
            if hasattr(self, setting):
                self.__setattr__(setting, False)

    def _fetch(self) -> dict:
        response = projects.get_project_settings(
            self._connection, self._project_id, whitelist=[('ERR001', 404)]
        )
        settings = response.json() if response.ok else {}
        if not response.ok:
            msg = (
                "Settings could not be fetched. It may be because the project is not "
                "loaded in the Intelligence Server or the project is idle."
            )
            helper.exception_handler(msg, Warning)
        return self._prepare_settings_fetch(settings)

    def _get_config(self):
        if not ProjectSettings._CONFIG:
            project_id = self._project_id
            if not project_id:
                project_id = Project._list_loaded_projects(
                    self._connection, to_dictionary=True
                )['id'][0]
            response = projects.get_project_settings_config(
                self._connection, project_id
            )
            ProjectSettings._CONFIG = response.json()
            super()._get_config()

    def _check_params(self, project_id: str | None = None):
        if project_id:
            super(BaseSettings, self).__setattr__('_project_id', project_id)
        if not self._connection or not self._project_id:
            raise AttributeError(
                "Please provide `connection` and `project_id` parameter"
            )

    @wip()
    def list_caching_properties(self, show_description: bool = False) -> dict:
        """
        Fetch current project settings connected with caching from I-Server

        Args:
            show_description (bool): if True return, description and value
                for each setting, else, return values only
        """
        self.fetch()
        return {
            k: v
            for (k, v) in self.list_properties(
                show_description=show_description
            ).items()
            if any(word in k.lower() for word in ("cache", "caching"))
        }



from mstrio.distribution_services.subscription.subscription_manager import (
    list_subscriptions,
)
from mstrio.server.language import Language
from mstrio.users_and_groups.user import list_users
from mstrio.connection import get_connection

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

ALTERED_LANGUAGE = Language(conn, id='000004124F95EF3956E52781700C1E7A')

# list active users
active_users = list_users(conn, enabled=True)

# set up default language for those users
for user in active_users:
    user.alter(language=ALTERED_LANGUAGE)

# list all inactive users and count them
inactive_users = list_users(conn, enabled=False)
inactive_users_count = len(inactive_users)
print('Number of inactive users:', inactive_users_count)

# list all related subscriptions for inactive users
for user in inactive_users:
    subscriptions = list_subscriptions(conn, project_id=conn.project_id, owner=user)
    print('Subscriptions for user', user.name, ':', subscriptions)



from enum import Enum


# A sentinel object to detect if a attribute has been returned by REST API
class _MissingType:
    pass


MISSING = _MissingType()


class ObjectTypes(Enum):
    FILTER = 1
    REPORT_DEFINITION = 3
    METRIC = 4
    AGG_METRIC = 7
    FOLDER = 8
    SUBSCRIPTION_DEVICE = 9
    PROMPT = 10
    FUNCTION = 11
    ATTRIBUTE = 12
    FACT = 13
    DIMENSION = 14
    TABLE = 15
    SHORTCUT_TYPE = 18
    MONITOR = 20
    ATTRIBUTE_FORM = 21
    COLUMN = 26
    PROPERTY_SET = 28
    DBROLE = 29
    DBLOGIN = 30
    DBCONNECTION = 31
    PROJECT = 32
    USER = 34
    USERGROUP = 34
    SUBSCRIPTION_TRANSMITTER = 35
    CONFIGURATION = 36
    SEARCH = 39
    ROLE = 43
    SECURITY_ROLE = 44
    LOCALE = 45
    CONSOLIDATION = 47
    SCHEDULE_EVENT = 49
    SCHEDULE_OBJECT = 50
    SCHEDULE_TRIGGER = 51
    DBTABLE = 53
    DOCUMENT_DEFINITION = 55
    DRILL_MAP = 56
    DBMS = 57
    SECURITY_FILTER = 58
    SHORTCUT = 67
    SHORTCUT_TARGET = 68
    DRIVER = 84
    NOT_SUPPORTED = None

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value

    @classmethod
    def _missing_(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        member._name_ = f'UNDEFINED_TYPE:{value}'
        return member

    @staticmethod
    def contains(item):
        if isinstance(item, ObjectTypes):
            return True
        return item in [v.value for v in ObjectTypes.__members__.values()]


class ObjectSubTypes(Enum):
    FILTER = 256
    REPORT_GRID = 768
    REPORT_GRAPH = 769
    REPORT_ENGINE = 770
    REPORT_GRID_AND_GRAPH = 774
    OLAP_CUBE = 776
    INCREMENTAL_REFRESH_REPORT = 777
    REPORT_TRANSACTION = 778
    SUPER_CUBE = 779
    REPORT_HYPER_CARD = 781
    FUNCTION = 2816
    USER = 8704
    USER_GROUP = 8705
    SEARCH = 9984
    FUNCTION_PACKAGE_DEFINITION = 10752
    REPORT_WRITING_DOCUMENT = 14081
    MD_SECURITY_FILTER = 14848
    ATTRIBUTE = 3072
    ATTRIBUTE_ABSTRACT = 3075
    ATTRIBUTE_FORM = 5376
    ATTRIBUTE_RECURSIVE = 3076
    ATTRIBUTE_ROLE = 3073
    ATTRIBUTE_TRANSFORMATION = 3074
    ROLE_TRANSFORMATION = 11009
    NOT_SUPPORTED = None

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value


class ExtendedType(Enum):
    RESERVED = 0
    RELATIONAL = 1
    MDX = 2
    CUSTOM_SQL_FREE_FORM = 3
    CUSTOM_SQL_WIZARD = 4
    DATA_IMPORT_GENERAL = 256
    DATA_IMPORT_FILE_EXCEL = 272
    DATA_IMPORT_FILE_TEXT = 288
    DATA_IMPORT_CUSTOM_SQL = 304
    DATA_IMPORT_TABLE = 320
    DATA_IMPORT_OAUTH = 336
    DATA_IMPORT_OAUTH_SFDC = 337
    DATA_IMPORT_OAUTH_GDRIVE = 338
    DATA_IMPORT_OAUTH_DROPBOX = 339
    _ = 341
    DATA_IMPORT_CUSTOM_SQL_WIZARD = 352
    DATA_IMPORT_SPARK = 416

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value


TypeOrSubtype = (
    int | ObjectTypes | ObjectSubTypes | list[int | ObjectTypes | ObjectSubTypes]
)



import functools
import logging

from requests_futures.sessions import FuturesSession


class FuturesSessionWithRenewal(FuturesSession):
    def __init__(self, *, connection, **kwargs):
        super().__init__(session=connection._session, **kwargs)
        self.connection = connection

    def request(self, *args, **kwargs):
        if self.connection._is_session_expired():
            self.connection._renew_or_reconnect()

        return super().request(*args, **kwargs)

    def get(self, *, endpoint, **kwargs):
        r"""
        Sends a GET request. Returns :class:`Future` object.

        :param endpoint: endpoint that will be combined with
            `self.connection.base_url`for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype : concurrent.futures.Future
        """
        url = self.connection.base_url + endpoint

        return super(FuturesSession, self).get(url, **kwargs)

    def options(self, *, endpoint, **kwargs):
        r"""Sends a OPTIONS request. Returns :class:`Future` object.

        :param endpoint: endpoint that will be combined with
            `self.connection.base_url`for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype : concurrent.futures.Future
        """
        url = self.connection.base_url + endpoint

        return super(FuturesSession, self).options(url, **kwargs)

    def head(self, *, endpoint, **kwargs):
        r"""Sends a HEAD request. Returns :class:`Future` object.

        :param endpoint: endpoint that will be combined with
            `self.connection.base_url`for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype : concurrent.futures.Future
        """
        url = self.connection.base_url + endpoint

        return super(FuturesSession, self).head(url, **kwargs)

    def post(self, *, endpoint, data=None, json=None, **kwargs):
        r"""Sends a POST request. Returns :class:`Future` object.

        :param endpoint: endpoint that will be combined with
            `self.connection.base_url`for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of
            the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype : concurrent.futures.Future
        """
        url = self.connection.base_url + endpoint

        return super(FuturesSession, self).post(url, data=data, json=json, **kwargs)

    def put(self, *, endpoint, data=None, **kwargs):
        r"""Sends a PUT request. Returns :class:`Future` object.

        :param endpoint: endpoint that will be combined with
            `self.connection.base_url`for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype : concurrent.futures.Future
        """
        url = self.connection.base_url + endpoint

        return super(FuturesSession, self).put(url, data=data, **kwargs)

    def patch(self, *, endpoint, data=None, **kwargs):
        r"""Sends a PATCH request. Returns :class:`Future` object.

        :param endpoint: endpoint that will be combined with
            `self.connection.base_url`for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype : concurrent.futures.Future
        """
        url = self.connection.base_url + endpoint

        return super(FuturesSession, self).patch(url, data=data, **kwargs)

    def delete(self, *, endpoint, **kwargs):
        r"""Sends a DELETE request. Returns :class:`Future` object.

        :param endpoint: endpoint that will be combined with
            `self.connection.base_url`for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype : concurrent.futures.Future
        """
        url = self.connection.base_url + endpoint

        return super(FuturesSession, self).delete(url, **kwargs)


def renew_session(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if (
            not kwargs.pop('skip_expiration_check', False)
            and self._is_session_expired()
        ):
            self._renew_or_reconnect()

        return func(self, *args, **kwargs)

    return wrapper


def log_request(logger):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, method_name, url=None, endpoint=None, **kwargs):
            if logger.isEnabledFor(logging.DEBUG):
                log_url = url or self.base_url + endpoint
                logger.debug("method = %s url = '%s'", method_name, log_url)
                logger.debug("headers = %s", self._session.headers)
                logger.debug("headers additional = %s", kwargs.get('headers'))
                logger.debug("body = %s", kwargs.get('json'))
            return func(self, method_name, url=url, endpoint=endpoint, **kwargs)

        return wrapper

    return decorator



# flake8: noqa

# isort: off
from .dbms import Dbms, list_available_dbms

# isort: on
from .database_connections import DatabaseConnections
from .datasource_connection import (
    CharEncoding,
    DatasourceConnection,
    DriverType,
    ExecutionMode,
    list_datasource_connections,
)
from .embedded_connection import EmbeddedConnection
from .datasource_instance import (
    DatasourceInstance,
    DatasourceType,
    list_connected_datasource_instances,
    list_datasource_instances,
)
from .datasource_login import DatasourceLogin, list_datasource_logins
from .datasource_map import DatasourceMap, list_datasource_mappings
from .driver import Driver, list_drivers
from .gateway import Gateway, list_gateways
from .helpers import DBType, GatewayType



from mstrio.api import usergroups as usergroups_api


def update_user_group_info(connection, id, body):
    """Update user group by a specified ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user group
        body: dict with changes to be made

    Returns:
        dict with user group information
    """
    # We do not want information about members from patch as it is insufficient
    # to differentiate between users and user groups
    response = usergroups_api.update_user_group_info(
        connection=connection, id=id, body=body
    )
    result = response.json()
    if result.get('members'):
        result.pop('members')

    return result



"""This is the demo script to show how to manage translations. This script will
not work without replacing parameters with real values. Its basic goal is to
present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from io import StringIO
from mstrio.datasources import DatasourceInstance
from mstrio.object_management import list_translations, Translation
from mstrio.modeling.metric import Metric
from mstrio.server import Language

# Define a variable which can be later used in a script
PROJECT_NAME = 'MicroStrategy Tutorial'  # Project to connect to

# Create connection based on workstation data
CONN = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
METRIC_ID = 'F03528954D22B1C3454BC68978712F94'
OBJECT_TYPE = 4
LANG_LCID = 1040
LANG_2_LCID = 1053
LANG_ID = '000004104F95EF3956E52781700C1E7A'
LANG_2_ID = '0000041D4F95EF3956E52781700C1E7A'
LANGUAGE_OBJECT = Language(connection=CONN, id=LANG_ID)
LANGUAGE_OBJECT_2 = Language(connection=CONN, id=LANG_2_ID)
TARGET_ID = 'MTowOi1GMDM1Mjg5NTREMjJCMUMzNDU0QkM2ODk3ODcxMkY5NC00'
TRANSLATION_VALUE = 'Data dell ordine +2'
TRANSLATION_VALUE_ALTERED = 'Data dell ordine +2 altered'
TRANSLATION_VALUE_2 = 'Orderdatum +2'
TRANSLATION_VALUE_2_ALTERED = 'Orderdatum +2 altered'

# You can manage translations in 2 ways, through the Translation class that
# allows managing Translations on any object.
# Or through the translation methods added to all objects that support
# translations, that are supported by mstrio-py.

# Below are snippets on using the Translation class methods:

# Add translations to an Object

# The translations are specified with a list of helper classes
# Each helper class has to contain the target_language, value and target_id
# For target_language you can either specify:
# - Language class object
# - specific Language ID
# - specific Language lcid

Translation.add_translation(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    translations=[
        Translation.OperationData(
            target_language=LANG_LCID,
            value=TRANSLATION_VALUE,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_LCID,
            value=TRANSLATION_VALUE_2,
            target_id=TARGET_ID,
        ),
    ],
    project_id=CONN.project_id,
)

# Alter translations of an Object

# The translations are specified with a list of helper classes
# The helper class has identical structure to the ones provided when adding
Translation.alter_translation(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    translations=[
        Translation.OperationData(
            target_language=LANG_ID,
            value=TRANSLATION_VALUE_ALTERED,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_ID,
            value=TRANSLATION_VALUE_2_ALTERED,
            target_id=TARGET_ID,
        ),
    ],
    project_id=CONN.project_id,
)

# Remove translations from an Object

# The translations are specified with a list of helper classes
# The helper class for removing does not need to contain the value
Translation.remove_translation(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    translations=[
        Translation.OperationData(target_language=LANGUAGE_OBJECT, target_id=TARGET_ID),
        Translation.OperationData(
            target_language=LANGUAGE_OBJECT_2, target_id=TARGET_ID
        ),
    ],
    project_id=CONN.project_id,
)

# Listing translations for an object
list_translations(
    connection=CONN, id=METRIC_ID, object_type=OBJECT_TYPE, project_id=CONN.project_id
)

# You can also specify a list of languages for the listing function
# The function will automatically list translations for only those languages
# For the languages you have to provide a list that contains:
# - Language class objects
# - specific Language IDs
# - specific Language lcids
# The three types of language argument can be mixed in the list
list_translations(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    project_id=CONN.project_id,
    languages=[LANG_LCID, LANG_2_ID],
)

# Translations can be managed from an object as well
METRIC = Metric(connection=CONN, id=METRIC_ID)

# Add translations to the object
# Only the list of translation helper classes needs to be provided
METRIC.add_translation(
    translations=[
        Translation.OperationData(
            target_language=LANG_LCID,
            value=TRANSLATION_VALUE,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_LCID,
            value=TRANSLATION_VALUE_2,
            target_id=TARGET_ID,
        ),
    ],
)

# Alter translations of the object
# Only the list of translation helper classes needs to be provided
METRIC.alter_translation(
    translations=[
        Translation.OperationData(
            target_language=LANG_LCID,
            value=TRANSLATION_VALUE_ALTERED,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_LCID,
            value=TRANSLATION_VALUE_2_ALTERED,
            target_id=TARGET_ID,
        ),
    ],
)

# Remove translations from the object
# Only the list of translation helper classes needs to be provided
METRIC.remove_translation(
    translations=[
        Translation.OperationData(target_language=LANG_LCID, target_id=TARGET_ID),
        Translation.OperationData(target_language=LANG_2_LCID, target_id=TARGET_ID),
    ],
)

# List translations of the object
METRIC.list_translations()

# You can specify a languages list for the object as well
METRIC.list_translations(languages=[LANG_LCID, LANG_2_ID])

# Export translations to CSV
CSV_LIST = StringIO(Translation.to_csv_from_list(connection=CONN, object_list=[METRIC]))
# Export translations to CSV with extra columns
CSV_LIST_EXTRA_COLUMNS = Translation.to_csv_from_list(
    connection=CONN,
    object_list=[METRIC],
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
)
# Export translations to CSV with specific languages only
CSV_LIST_SPECIFIED_LANGUAGES = StringIO(
    Translation.to_csv_from_list(
        connection=CONN, object_list=[METRIC], languages=[LANG_ID, LANG_2_ID]
    )
)
# Export translations to CSV in denormalized format and specific separator
CSV_LIST_DENORMALIZED = Translation.to_csv_from_list(
    connection=CONN, object_list=[METRIC], denormalized_form=True, separator=','
)

# Add translations from CSV
Translation.add_translations_from_csv(connection=CONN, file_path=CSV_LIST)
# Add translations from CSV with removing translations and automatching target_ids
Translation.add_translations_from_csv(
    connection=CONN,
    file_path=CSV_LIST_SPECIFIED_LANGUAGES,
    delete=True,
    automatch_target_ids=True,
)

# Export translations to JSON
JSON_LIST = StringIO(
    Translation.to_json_from_list(connection=CONN, object_list=[METRIC])
)
# Export translations to JSON with extra columns
JSON_LIST_EXTRA_COLUMNS = Translation.to_json_from_list(
    connection=CONN,
    object_list=[METRIC],
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
)
# Export translations to JSON with specific languages only
JSON_LIST_SPECIFIED_LANGUAGES = StringIO(
    Translation.to_json_from_list(
        connection=CONN, object_list=[METRIC], languages=[LANG_ID, LANG_2_ID]
    )
)
# Export translations to JSON in denormalized format
JSON_LIST_DENORMALIZED = Translation.to_json_from_list(
    connection=CONN, object_list=[METRIC], denormalized_form=True
)

# Add translations from JSON
Translation.add_translations_from_json(connection=CONN, file_path=JSON_LIST)
# Add translations from JSON with removing translations and automatching target_ids
Translation.add_translations_from_json(
    connection=CONN,
    file_path=JSON_LIST_SPECIFIED_LANGUAGES,
    delete=True,
    automatch_target_ids=True,
)

# Export translations to database
# The flag force=True will be used in the examples
# That flag means no confirmation will be asked before overwriting the specified table
# If the flag is not used, the user will be asked for confirmation before overwriting
DATASOURCE = DatasourceInstance(connection=CONN, id='A23BBC514D336D5B4FCE919FE19661A3')
Translation.to_database_from_list(
    connection=CONN,
    object_list=[METRIC],
    table_name='test_table',
    datasource=DATASOURCE,
    force=True,
)
# Export translations to database with extra columns
Translation.to_database_from_list(
    connection=CONN,
    object_list=[METRIC],
    table_name='test_table',
    datasource=DATASOURCE,
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
    force=True,
)
# Export translations to database with specific languages only
Translation.to_database_from_list(
    connection=CONN,
    object_list=[METRIC],
    table_name='test_table',
    datasource=DATASOURCE,
    languages=[LANG_ID, LANG_2_ID],
    force=True,
)
# Export translations to database in denormalized format
Translation.to_database_from_list(
    connection=CONN,
    object_list=[METRIC],
    table_name='test_table_denormalized',
    datasource=DATASOURCE,
    denormalized_form=True,
    force=True,
)

# Add translations from database
Translation.add_translations_from_database(
    connection=CONN, table_name='test_table', datasource=DATASOURCE
)
# Add translations from database with removing translations and automatching target_ids
Translation.add_translations_from_database(
    connection=CONN,
    table_name='test_table',
    datasource=DATASOURCE,
    delete=True,
    automatch_target_ids=True,
)

# Export translations to dataframe
DATAFRAME = Translation.to_dataframe_from_list(connection=CONN, object_list=[METRIC])
# Export translations to dataframe with extra columns
DATAFRAME_EXTRA_COLUMNS = Translation.to_dataframe_from_list(
    connection=CONN,
    object_list=[METRIC],
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
)
# Export translations to dataframe with specific languages only
DATAFRAME_SPECIFIED_LANGUAGES = Translation.to_dataframe_from_list(
    connection=CONN, object_list=[METRIC], languages=[LANG_ID, LANG_2_ID]
)
# Export translations to dataframe in denormalized format
DATAFRAME_DENORMALIZED = Translation.to_dataframe_from_list(
    connection=CONN, object_list=[METRIC], denormalized_form=True
)

# Add translations from dataframe
Translation.add_translations_from_dataframe(connection=CONN, dataframe=DATAFRAME)
# Add translations from dataframe with removing translations and automatching target_ids
Translation.add_translations_from_dataframe(
    connection=CONN, dataframe=DATAFRAME, delete=True, automatch_target_ids=True
)



from mstrio.connection import get_connection
from mstrio.users_and_groups.user import User, list_users
from mstrio.users_and_groups.user_group import UserGroup

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# Move users from one user group to another
user_group_1 = UserGroup(conn, name='Second Factor Exempt')
user_group_2 = UserGroup(conn, name='TransactionServer')
user_group_2.add_users([user['id'] for user in user_group_1.members])
user_group_1.remove_all_users()

# change owner for all the users to Administrator
for user in list_users(conn):
    user.alter(owner=User(conn, name='Administrator'))

# change/add LDAP Distinguished Name (DN) on target user group
user_group = UserGroup(conn, name='Second Factor Exempt')
user_group.alter(ldapdn='CN=Users,DC=domain,DC=com')



"""This is the demo script to show how administrator can manage VLDB settings.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.server import Project
from mstrio.project_objects.datasets import OlapCube
from mstrio.datasources import DatasourceInstance
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project to interact with

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Project VLDB settings management
# Get project with a given name
project = Project(connection=conn, name=PROJECT_NAME)

# Define variables which can be later used in a script
PROPERTY_SET = $property_set    # Insert name of property set
GROUP_NAME = $group_name        # Insert name of group
GROUP_ID = $group_id            # Insert ID of group
SETTING_NAME1 = $setting_name1  # Insert name of first VLDB setting
SETTING_NAME2 = $setting_name2  # Insert name of second VLDB setting
SETTING_NAME3 = $setting_name3  # Insert name of third VLDB setting

# Listing VLDB setting properties
vldb_setting_properties = project.vldb_settings[SETTING_NAME1].list_properties()
print(vldb_setting_properties)

# Listing VLDB settings with different conditions
project_settings = project.list_vldb_settings()
print(project_settings)
project_settings_as_dicts = project.list_vldb_settings(to_dictionary=True)
print(project_settings_as_dicts)
project_settings_as_dataframe = project.list_vldb_settings(to_dataframe=True)
print(project_settings_as_dataframe)
project_settings_from_set = project.list_vldb_settings(set_names=[PROPERTY_SET])
print(project_settings_from_set)
project_settings_from_group_by_name = project.list_vldb_settings(groups=[GROUP_NAME])
print(project_settings_from_group_by_name)
project_settings_from_group_by_id = project.list_vldb_settings(groups=[GROUP_ID])
print(project_settings_from_group_by_id)
project_settings_from_names = project.list_vldb_settings(
    names=[SETTING_NAME1, SETTING_NAME2, SETTING_NAME3])
print(project_settings_from_names)

# Define variables which can be later used in a script
SETTING_VALUE1 = $setting_value1  # Insert new value for first VLDB setting
SETTING_VALUE2 = $setting_value2  # Insert new value for second VLDB setting
SETTING_VALUE3 = $setting_value3  # Insert new value for third VLDB setting

# Altering different VLDB settings
project.reset_vldb_settings(force=True)
print(project.vldb_settings[SETTING_NAME1].value)
print(project.vldb_settings[SETTING_NAME2].value)
print(project.vldb_settings[SETTING_NAME3].value)
project.alter_vldb_settings(names_to_values={
    SETTING_NAME1: SETTING_VALUE1,
    SETTING_NAME2: SETTING_VALUE2,
    SETTING_NAME3: SETTING_VALUE3
})
print(project.vldb_settings[SETTING_NAME1].value)
print(project.vldb_settings[SETTING_NAME2].value)
print(project.vldb_settings[SETTING_NAME3].value)

# Reseting VLDB settings with different conditions
# Set values to default for all settings in set
project.reset_vldb_settings(set_names=[PROPERTY_SET])
project_settings_from_set = project.list_vldb_settings(set_names=[PROPERTY_SET])
# Set values to default for all settings in group specified by name
project.reset_vldb_settings(groups=[GROUP_NAME])
project_settings_from_group_by_name = project.list_vldb_settings(groups=[GROUP_NAME])
# Set values to default for all settings in group specified by ID
project_settings_from_group_by_id = project.list_vldb_settings(groups=[GROUP_ID])
# Set values to default for all settings specified by names
project.reset_vldb_settings(names=[SETTING_NAME1, SETTING_NAME2, SETTING_NAME3])
project_settings_from_names = project.list_vldb_settings(
    names=[SETTING_NAME1, SETTING_NAME2, SETTING_NAME3])
print(project.vldb_settings[SETTING_NAME1].value)
print(project.vldb_settings[SETTING_NAME2].value)
print(project.vldb_settings[SETTING_NAME3].value)

# Datasource VLDB settings management

# Define a variable which can be later used in a script
DATASOURCE_NAME = $datasource_name  # Insert name of datasource to interact with

# Get datasource with a given name
datasource = DatasourceInstance(conn, name=DATASOURCE_NAME)

# Listing VLDB settings with different conditions
datasource_settings_as_dicts_from_set = datasource.list_vldb_settings(
    to_dictionary=True, set_names=[PROPERTY_SET])
print(datasource_settings_as_dicts_from_set)
datasource_settings_as_dataframe_from_group = datasource.list_vldb_settings(
    to_dataframe=True, groups=[GROUP_NAME])
print(datasource_settings_as_dataframe_from_group)
datasource_settings_from_names = datasource.list_vldb_settings(
    names=[SETTING_NAME1, SETTING_NAME2])
print(datasource_settings_from_names)

# Altering different VLDB settings
datasource.reset_vldb_settings(force=True)
print(datasource.vldb_settings[SETTING_NAME1].value)
print(datasource.vldb_settings[SETTING_NAME2].value)
datasource.alter_vldb_settings(names_to_values={
    SETTING_NAME1: SETTING_VALUE1,
    SETTING_NAME2: SETTING_VALUE2
})
print(datasource.vldb_settings[SETTING_NAME1].value)
print(datasource.vldb_settings[SETTING_NAME2].value)

# Reseting VLDB settings with different conditions
# Set values to default for all settings in set
datasource.reset_vldb_settings(set_names=[PROPERTY_SET])
# Set values to default for all settings specified by names
datasource.reset_vldb_settings(names=[SETTING_NAME2, SETTING_NAME3])
datasource_settings_from_names = datasource.list_vldb_settings(
    names=[SETTING_NAME1, SETTING_NAME2, SETTING_NAME3])
print(datasource_settings_from_names[SETTING_NAME1].value)
print(datasource_settings_from_names[SETTING_NAME2].value)
print(datasource_settings_from_names[SETTING_NAME3].value)

# OLAP Cube VLDB settings management

# Define a variable which can be later used in a script
CUBE_ID = $cube_id

# Get OLAP Cube based on its ID
olap_cube = OlapCube(connection=conn, id=CUBE_ID)

# Listing VLDB settings with different conditions
cube_settings_from_set = olap_cube.list_vldb_settings(set_names=[PROPERTY_SET])
print(cube_settings_from_set)
cube_settings_as_dicts = olap_cube.list_vldb_settings(to_dictionary=True)
print(cube_settings_as_dicts)
cube_settings_as_dataframe_from_group = olap_cube.list_vldb_settings(
    to_dataframe=True, groups=[GROUP_NAME]
)
print(cube_settings_as_dataframe_from_group)

# Altering different VLDB settings
olap_cube.reset_vldb_settings(force=True)
print(olap_cube.vldb_settings[SETTING_NAME3].value)
olap_cube.alter_vldb_settings(names_to_values={SETTING_NAME3: SETTING_VALUE3})
print(olap_cube.vldb_settings[SETTING_NAME3].value)

# Reseting VLDB settings
# Set values to default for all settings specified by names
olap_cube.reset_vldb_settings(names=[SETTING_NAME3])
cube_settings_from_names = olap_cube.list_vldb_settings(names=[SETTING_NAME3])
print(cube_settings_from_names[SETTING_NAME3].value)



import logging
from typing import Any

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.server.cluster import Cluster
from mstrio.utils import helper
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


@class_version_handler('11.2.0000')
class DatabaseConnections:
    """Browse and manage database connections on the environment.

    Attributes:
        connection: A MicroStrategy connection object
    """

    def __init__(self, connection: Connection):
        """Initialize the `DatabaseConnections` object.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
        """
        self.connection = connection

    def list_connections(
        self, nodes: str | list[str] = None, limit: int | None = None, **filters
    ) -> list[dict[str, Any]]:
        """Get all active database connections. Optionally filter the
         connections by specifying the `filters` keyword arguments.

        Args:
            nodes: Node (server) names on which databases will be disconnected.
            limit: limit the number of elements returned. If `None`, all objects
                are returned.
            **filters: Available filter parameters: ['status',
                'database_instance_name', 'database_instance_id', 'user_name',
                'database_login_name', 'cluster_node', 'id', 'name', 'type']
        """
        if nodes is None:
            all_nodes = Cluster(self.connection).list_nodes(to_dictionary=True)
            nodes = [n['name'] for n in all_nodes if n['status'] == 'running']

        nodes_names = ",".join(nodes) if isinstance(nodes, list) else nodes
        all_databases = helper.fetch_objects(
            connection=self.connection,
            api=monitors.get_database_connections,
            nodes_names=nodes_names,
            dict_unpack_value="dbConnectionInstances",
            limit=limit,
            filters=filters,
        )
        return all_databases

    def disconnect_database(self, connection_id: str, force: bool = False) -> bool:
        """Disconnect database connections by passing in connection_id.

        Args:
            connection_id: Database Connection Instance Id
            force: if True, no additional prompt will be shown before.
                Default False.
            Returns:
                True for success. False otherwise.
        """
        user_input = 'N'
        if not force:
            user_input = input(
                f"Are you sure you want to disconnect database connection "
                f"'with ID:{connection_id}? [Y/N]: "
            )
        if force or user_input == 'Y':
            response = monitors.delete_database_connection(
                self.connection, connection_id
            )
            if response.status_code == 204 and config.verbose:
                logger.info(
                    f'Successfully disconnected database connection instance '
                    f'{connection_id}.'
                )
            return response.ok
        else:
            return False

    def disconnect_all_databases(self, force: bool = False) -> list[dict] | None:
        """Disconnect all database connections.

        Args:
            force: if True, no additional prompt will be shown before
                disconnecting all connections

        Returns:
            - list of statuses of disconnecting all connections with their ids
               and messages from the I-Server
            - in case of error it returns None
        """
        if not force:
            user_input = input(
                "Are you sure you want to disconnect all database connections? [Y/N]: "
            )
            if user_input != "Y":
                return None
            else:
                force = True

        connections = self.list_connections()
        threads = helper.get_parallel_number(len(connections))
        with FuturesSessionWithRenewal(
            connection=self.connection, max_workers=threads
        ) as session:
            futures = [
                monitors.delete_database_connection_async(session, conn["id"])
                for conn in connections
            ]
            statuses: list[dict[str, str | int]] = []
            for f in futures:
                response = f.result()
                statuses.append(
                    {
                        'id': response.url.rsplit("/").pop(-1),
                        'status': response.status_code,
                    }
                )
        return self._prepare_disconnect_by_id_message(statuses=statuses)

    @staticmethod
    def _prepare_disconnect_by_id_message(statuses: list[dict]) -> list[dict]:
        succeeded = []
        failed = []

        for s in statuses:
            status = s['status']
            if status < 400:
                succeeded.append(s['id'])
            else:
                # in most cases there will be status code 403
                failed.append(s['id'])

        if config.verbose:
            if succeeded:
                logger.info(
                    'Database connections with ids listed below were successfully '
                    'disconnected:\n\t' + ',\n\t'.join(succeeded)
                )
            if failed:
                logger.warning(
                    'Database connections with ids listed below were not '
                    'disconnected:\n\t' + ',\n\t'.join(failed)
                )
        return statuses



from __future__ import annotations

import logging
from dataclasses import dataclass

from mstrio import config
from mstrio.connection import Connection
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.utils.entity import DeleteMixin, Entity
from mstrio.utils.helper import Dictable, delete_none_values
from mstrio.utils.response_processors import languages, objects
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler(version='11.3.1060')
def list_languages(
    connection: Connection,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list[Language] | list[dict]:
    """Get all languages as a list of Language objects or dictionaries.
    Optionally filter the languages by specifying filters.

    Args:
        connection: MicroStrategy connection object
        to_dictionary: if True returns a list of Language dicts,
            otherwise returns a list of Language objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters:
            ['name', 'base_language_lcid', 'lcid', 'interface_language',
            'last_modified', 'hidden', 'is_language_supported']
    """
    return Language._list_languages(
        connection=connection, to_dictionary=to_dictionary, limit=limit, **filters
    )


@method_version_handler(version='11.3.1060')
def list_interface_languages(
    connection: Connection, to_dictionary: bool = False
) -> list[Language.InterfaceLanguage] | list[dict]:
    """List all available interface languages.

    Args:
        connection (Connection): MicroStrategy connection object returned
            by `connection.Connection()`
        to_dictionary: if True returns a list of InterfaceLanguage dicts,
            otherwise returns a list of InterfaceLanguage objects

    Returns:
    A list of interface languages as dictionaries."""
    interface_languages = languages.get_interface_languages(connection=connection)
    if to_dictionary:
        return interface_languages
    else:
        return [
            Language.InterfaceLanguage.from_dict(source=obj)
            for obj in interface_languages
        ]


@class_version_handler(version='11.3.1060')
class Language(Entity, DeleteMixin, TranslationMixin):
    """Python representation of a Microstrategy Language object.

    Attributes:
        id (str): language's ID
        name (str): language's name
        base_language_lcid (int): LC ID of the language serving as a base for
            the language
        lcid (int): LC ID of the language
        owner (User): owner of the language
        last_modified (datetime): date of when language was last modified
        formatting_settings (TimeInterval): formatting settings of the language
        interface_language (InterfaceLanguage): details of the language's
            interface language
        hidden (bool): whether the language is hidden
        is_language_supported (bool): whether the language is supported
    """

    @dataclass
    class TimeInterval(Dictable):
        """Object that stores formatting settings of the Language.

        Attributes:
            minutes15 (str): formatting settings for 15 minutes
            minutes30 (str): formatting settings for 30 minutes
            hour (str): formatting settings for an hour
            day (str): formatting settings for a day
            week (str): formatting settings for a week
            hour_of_day (str): formatting settings for the hour of the day
            month (str): formatting settings for a month
            quarter (str): formatting settings for a quarter
            year (str): formatting settings for a year

        Formatting settings instructions:
            To specify formatting for:
                hours: use the letter h
                minutes: use the letter m
                days: use the letter d
                months: use the capital letter M
                quarters: use the letter q
                years: use the letter y
            To specify 12 hour clock use AM/PM afterwards, example: h:mm AM/PM

         Default formatting settings examples:
            minutes15: h:mm AM/PM -> 9:44 AM
            minutes30: h:mm AM/PM -> 9:44 AM
            hour: h:mm AM/PM -> 9:44 AM
            day: M/d -> 7/26
            week: M/d -> 7/26
            hour_of_day: h:mm AM/PM -> 9:44 AM
            month: MMM -> Jul
            quarter: qqq -> Q3
            year: yyyy -> 2023"""

        minutes15: str
        minutes30: str
        hour: str
        day: str
        week: str
        hour_of_day: str
        month: str
        quarter: str
        year: str

    @dataclass
    class InterfaceLanguage(Dictable):
        """Object that stores the interface language of the Language.

        Attributes:
            name (str): name of the interface language
            id (str): id of the interface language
            sub_type (ObjectSubTypes): subtype of the interface language,
                defaults to None
        """

        name: str
        id: str
        sub_type: ObjectSubTypes | None = None

    _OBJECT_TYPE = ObjectTypes.LOCALE
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'base_language_lcid',
            'lcid',
            'interface_language',
            'last_modified',
            'hidden',
            'is_language_supported',
        ): languages.get,
        (
            'abbreviation',
            'type',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
        ): objects.get_info,
    }

    @staticmethod
    def _parse_owner(source, connection, to_snake_case: bool = True):
        """Parses owner from the API response."""
        from mstrio.users_and_groups import User

        return User.from_dict(source, connection, to_snake_case)

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': _parse_owner,
        'formatting_settings': TimeInterval.from_dict,
        'interface_language': InterfaceLanguage.from_dict,
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
    ) -> None:
        """Initializes a new instance of a Language class

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, Optional): Language's ID, defaults to None
            name (str, Optional): Language's name, defaults to None

        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Raises:
            ValueError: if both `id` and `name` are not provided
                or if Language with the given `name` doesn't exist.
        """
        if not id:
            if name:
                language = self.__find_language_by_name(
                    connection=connection, name=name
                )
                id = language['id']
            else:
                raise ValueError("Please provide either 'id' or 'name' argument.")
        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.name = kwargs.get('name')
        self.base_language_lcid = kwargs.get('base_laguage_lcid')
        self.lcid = kwargs.get('lcid')
        self.interface_language = (
            Language.InterfaceLanguage.from_dict(inl)
            if (inl := kwargs.get('interface_language'))
            else None
        )
        self.last_modified = kwargs.get('last_modified')
        self._hidden = kwargs.get('hidden')
        self._formatting_settings = None
        self.is_language_supported = kwargs.get('is_language_supported')

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        base_language: Language | str | int,
        interface_language_id: str | None = None,
        formatting_settings: TimeInterval | None = None,
    ) -> Language:
        """Create a new language with specified properties.

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            name (str): the name for the new Language
            base_language (Language | str | int): one of the following:
                - lcid attribute of the language that will be used as a base
                    language for the new Language
                - ID of the language that will be used as a base language for
                    the new Language
                - Language class object that will be used as a base language for
                    the new Language
            interface_language_id (str, Optional): id of the new Language's
                interface language
            formatting_settings (TimeInterval, Optional): formatting settings
                for the new Language

        Returns:
            Language class object.
        """
        if isinstance(base_language, int):
            base_lcid = base_language
        elif isinstance(base_language, str):
            base_lcid = Language(connection=connection, id=base_language).lcid
        elif isinstance(base_language, Language):
            base_lcid = base_language.lcid
        else:
            raise ValueError("Please provide a valid base language.")
        body = {
            'name': name,
            'baseLanguageLcid': base_lcid,
            'interfaceLanguageId': interface_language_id,
        }
        body = delete_none_values(source=body, recursion=True)
        language = cls.from_dict(
            source=languages.create(connection=connection, body=body),
            connection=connection,
        )
        if config.verbose:
            logger.info(
                f"Successfully created Language named: '{name}' with ID:"
                f" '{language.id}'"
            )
        if formatting_settings:
            language._formatting_settings = language._update_formatting_settings(
                formatting_settings=formatting_settings
            )
        else:
            language._formatting_settings = Language.TimeInterval.from_dict(
                source=languages.get_formatting_settings(
                    connection=connection, id=language.id
                )
            )
        return language

    def alter(
        self,
        name: str | None = None,
        formatting_settings: TimeInterval | None = None,
    ) -> None:
        """Alter the language's specified properties.

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            name (str, Optional): new name for the Language
            formatting_settings (TimeInverval, Optional): new formatting
                settings for the Language"""
        if name:
            self._alter_properties(name=name)
        if formatting_settings:
            self._formatting_settings = self._update_formatting_settings(
                formatting_settings=formatting_settings
            )

    def _update_formatting_settings(
        self, formatting_settings: TimeInterval
    ) -> TimeInterval:
        """Updates the formatting settings of the Language.

        Args:
            formatting_settings (TimeInterval): new formatting settings

        Returns:
            New formatting settings as TimeInterval class object."""
        formatting_body = {
            'operationList': [
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/minutes15',
                    'value': formatting_settings.minutes15,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/minutes30',
                    'value': formatting_settings.minutes30,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/hour',
                    'value': formatting_settings.hour,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/day',
                    'value': formatting_settings.day,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/week',
                    'value': formatting_settings.week,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/hourOfDay',
                    'value': formatting_settings.hour_of_day,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/month',
                    'value': formatting_settings.month,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/quarter',
                    'value': formatting_settings.quarter,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/year',
                    'value': formatting_settings.year,
                },
            ]
        }
        return Language.TimeInterval.from_dict(
            languages.update_formatting_settings(
                connection=self.connection, id=self.id, body=formatting_body
            )
        )

    def _remove_interface_language(self) -> None:
        """Removes the interface language from the Language if one exists."""
        if self.interface_language:
            op_list = [
                {
                    'op': 'remove',
                    'path': '/interfaceLanguage',
                    'value': self.interface_language.id,
                }
            ]
            languages.update(
                connection=self.connection, id=self.id, body={'operationList': op_list}
            )
            self.interface_language = None

    @classmethod
    def _list_languages(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list[Language] | list[dict]:
        objects = languages.get_all(connection=connection, limit=limit, filters=filters)
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @staticmethod
    def __find_language_by_name(connection: Connection, name: str):
        languages = list_languages(connection=connection, name=name)

        if languages:
            number_of_languages = len(languages)
            if number_of_languages > 1:
                raise ValueError(
                    f"There are {number_of_languages} Languages"
                    " with this name. Please initialize with id."
                )
            else:
                return languages[0].to_dict()
        else:
            raise ValueError(f"There is no Language with the given name: '{name}'")

    @property
    def hidden(self):
        return self._hidden

    @property
    def formatting_settings(self):
        if self._formatting_settings:
            return self._formatting_settings
        else:
            self._formatting_settings = Language.TimeInterval.from_dict(
                source=languages.get_formatting_settings(
                    connection=self.connection, id=self.id
                )
            )



from enum import Enum

_COMMON_ENUM_FIELDS = {
    'SUBSCRIPTION_RECIPIENT_NAME': 'recipient_name',
    'SUBSCRIPTION_OWNER_NAME': 'owner_name',
    'SUBSCRIPTION_REPORT_OR_DOCUMENT_NAME': 'report_document_name',
    'SUBSCRIPTION_PROJECT_NAME': 'project_name',
    'SUBSCRIPTION_DELIVERY_METHOD': 'delivery_method',
    'SUBSCRIPTION_SCHEDULE': 'schedule',
    'SUBSCRIPTION_NAME': 'subscription_name',
    'DELIVERY_STATUS': 'delivery_status',
    'DELIVERY_DATE': 'date',
    'DELIVERY_TIME': 'time',
    'DELIVERY_ERROR_MESSAGE': 'error_message',
}

FailedEmailDelivery = Enum(
    'FailedEmailDelivery',
    {**_COMMON_ENUM_FIELDS, 'DELIVERY_EMAIL_ADDRESS': 'email_address'},
)

FailedFileDelivery = Enum(
    'FailedFileDelivery',
    {
        **_COMMON_ENUM_FIELDS,
        'FILE_LOCATION': 'file_location',
        'LINK_TO_FILE': 'link_to_file',
    },
)

FailedFTPDelivery = Enum(
    'FailedFTPDelivery',
    {
        **_COMMON_ENUM_FIELDS,
        'FILE_LOCATION': 'file_location',
        'LINK_TO_FILE': 'link_to_file',
    },
)

FailedPrinterDelivery = Enum(
    'FailedPrinterDelivery',
    {
        **_COMMON_ENUM_FIELDS,
        'PRINTER_NAME': 'printer_name',
    },
)

FailedHistoryListDelivery = Enum(
    'FailedHistoryListDelivery',
    {
        **_COMMON_ENUM_FIELDS,
        'LINK_TO_HISTORY_LIST': 'link_to_history_list',
    },
)

FailedMobileDelivery = Enum('FailedMobileDelivery', _COMMON_ENUM_FIELDS)

FailedCacheCreation = Enum('FailedCacheCreation', _COMMON_ENUM_FIELDS)


class CompressionLevel(Enum):
    """Level of File | Email | FTP delivery compression"""

    OFF = 0
    LOW = 3
    MEDIUM = 6
    HIGH = 9


class CacheEncryptionLevel(Enum):
    """Cache encryption level on disk"""

    NONE = 0
    LOW = 1
    HIGH = 2


class ShowBaseViewInLibrary(Enum):
    """Always open dashboards/dossiers/documents with the last saved view
    in Library"""

    DEFAULT = 'use_inherited_value'
    YES = 'yes'
    NO = 'no'


class SmartMemoryUsageForIntelligenceServer(Enum):
    """Smart Memory Usage for Intelligence Server"""

    USE_INHERITED_VALUE = -1
    APPLY_BEST_STRATEGY = 0
    TURN_OFF_THE_CAPABILITY_WITHOUT_EXCEPTIONS = 1
    DISABLE_THE_CAPABILITY = 2
    ENABLE_THE_CAPABILITY = 3


class OrderMultiSourceDBI(Enum):
    """Rules to order multi-source Database Instances"""

    MULTISOURCE_DEFAULT_ORDERING = 0
    PROJECT_LEVEL_DATABASE_ORDERING = 1


class DisplayEmptyReportMessageInRWD(Enum):
    """Display mode for empty Grid/Graphs in documents"""

    DISPLAY_MESSAGE_IN_DOCUMENT_GRIDS = -1
    HIDE_DOCUMENT_GRID = 0


class MergeSecurityFilters(Enum):
    """Rules to merge multiple Security Filters across user groups and users"""

    UNION = 0
    INTERSECT = 3



from mstrio.connection import get_connection

from mstrio.distribution_services.subscription.email_subscription import (
    EmailSubscription,
)
from mstrio.distribution_services.subscription.content import Content
from mstrio.distribution_services.schedule.schedule import Schedule
from mstrio.distribution_services.subscription.file_subscription import FileSubscription
from mstrio.distribution_services.subscription.history_list_subscription import (
    HistoryListSubscription,
)
from mstrio.distribution_services.subscription.mobile_subscription import (
    MobileSubscription,
)
from mstrio.distribution_services.subscription.delivery import ClientType
from mstrio.distribution_services.subscription.subscription_manager import (
    list_subscriptions,
)
from mstrio.project_objects.dashboard import Dashboard


conn = get_connection(workstationData, 'MicroStrategy Tutorial')

example_schedule1 = Schedule(conn, id="987DF6464625F087CCC67FA9DB53097D")
example_schedule2 = Schedule(conn, id="FF7BB3B311D501F0C00051916B98494F")

example_user_id = 'E96A7BE711D4BBCE10004694316DE8A4'


# Create and execute an Email Subscription for a Document
email_sub = EmailSubscription.create(
    connection=conn,
    name='<Name of the subscription>',
    project_name='MicroStrategy Tutorial',
    contents=Content(
        id='C68851F14333D950BC37EC9D8DF3FDCA',
        type=Content.Type.DOCUMENT,
        personalization=Content.Properties(
            format_type=Content.Properties.FormatType.HTML
        ),
    ),
    schedules=[example_schedule1, example_schedule2],
    recipients=[example_user_id],
    email_subject="<Subject of the email>",
)
email_sub.execute()


# Modify a File Subscription for a Document with Bursting option on
example_file_sub = FileSubscription(conn, id="<ID of the subscription>")
file_sub_content: Content = example_file_sub.contents[0]
file_sub_content.personalization.bursting = Content.Properties.Bursting(
    slicing_attributes=['<Attribute ID>', '<Attribute ID>'],
)
example_file_sub.alter(content=file_sub_content)


# Create and execute a History List subscription for an OLAP Cube
example_hl_sub = HistoryListSubscription.create(
    connection=conn,
    name='<Name of the subscription>',
    contents=Content(
        id='C68851F14333D950BC37EC9D8DF3FDCA',
        type=Content.Type.CUBE,
        personalization=Content.Properties(
            format_type=Content.Properties.FormatType.HTML
        ),
    ),
    schedules=[example_schedule1, example_schedule2],
    recipients=[example_user_id],
    allow_delivery_changes=False,
    allow_personalization_changes=False,
    allow_unsubscribe=False,
)
example_hl_sub.execute()

# Create and execute a Mobile Subscription for a Document
example_mobile_sub = MobileSubscription.create(
    connection=conn,
    name='<Name of the subscription>',
    contents=Content(
        id='C68851F14333D950BC37EC9D8DF3FDCA',
        type=Content.Type.DOCUMENT,
        personalization=Content.Properties(
            format_type=Content.Properties.FormatType.HTML
        ),
    ),
    schedules=[example_schedule1, example_schedule2],
    recipients=[example_user_id],
    mobile_client_type=ClientType.PHONE,
    device_id='E206C75BABC441C5B13B60C5D956F605',
    do_not_create_update_caches=True,
    overwrite_older_version=False,
    re_run_hl=True,
)
example_mobile_sub.execute()


# Select a Subscription and list its dependent Dashboards
example_subscription = HistoryListSubscription(
    connection=conn, id="7138EE3011EA0FD26B470080EF752E56"
)
all_contents = example_subscription.contents
dashboard_contents = [
    content for content in all_contents if content.type == Content.Type.DASHBOARD
]

dashboard_ids = [content.id for content in dashboard_contents]
print("Dashboard IDs:")
print(dashboard_ids)

dashboards_objects = [Dashboard(conn, id) for id in dashboard_ids]
print(dashboards_objects)


# Select a Dashboard and find all Subscriptions using it. This step requires
# filtering a list of Subscriptions with Python list operations,
# as Document/Dashboard query responses do not backlink to Subscriptions.
all_subscriptions = list_subscriptions(conn)
dashboard_id = "<ID of the Dashboard>"

subscriptions_using_dashboard = [
    sub
    for sub in all_subscriptions
    if any(content.id == dashboard_id for content in sub.contents)
]
print("Subscriptions using the Dashboard:")
print(subscriptions_using_dashboard)



"""This is the demo script to show how administrator can manage contacts
 and contact addresses.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.distribution_services import Device
from mstrio.users_and_groups import (
    Contact, ContactAddress, ContactDeliveryType, ContactGroup, list_contacts, User
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project name here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# get list of all contacts
contacts = list_contacts(conn)

# get limited number of contacts
contacts = list_contacts(conn, limit=1)

# Define variables which can be later used in a script
CONTACT_ID = $contact_id  # Insert your contact ID here
CONTACT_NAME = $contact_name  # Insert your contact name here

# get contact by id
contact = Contact(conn, id=CONTACT_ID)

# get contact by name
contact = Contact(conn, name=CONTACT_NAME)

# Define variables which can be later used in a script
CONTACT_ADDRESS_NAME = $contact_address_name  # Insert your contact address name here
PHYSICAL_ADDRESS = $physical_address  # Insert physical address of created contact here
DEVICE_ID = $device_id  # Insert device used for contact here
CONTACT_DESCRIPTION = $contact_description  # Insert  description for new contact here
USERNAME = $username  # Insert name of user to link to the contact

# create contact address by providing device object
contact_address = ContactAddress(
    name=CONTACT_ADDRESS_NAME,
    physical_address=PHYSICAL_ADDRESS,
    # the ContactDeliveryType values can be found in distribution_services/contact/contact.py
    delivery_type=ContactDeliveryType.EMAIL,
    device=Device(conn, id=DEVICE_ID),
    is_default=True
)

# create contact address by providing device id
contact_address = ContactAddress(
    name=CONTACT_ADDRESS_NAME,
    physical_address=PHYSICAL_ADDRESS,
    # the ContactDeliveryType values can be found in distribution_services/contact/contact.py
    delivery_type=ContactDeliveryType.EMAIL,
    device=DEVICE_ID,
    is_default=True,
    connection=conn
)

# create contact
new_contact = Contact.create(
    connection=conn,
    name=CONTACT_NAME,
    linked_user=User(conn, name=USERNAME),
    description=CONTACT_DESCRIPTION,
    enabled=True,
    contact_addresses=[contact_address] # list of `ContactAddress` objects
)

# list properties for contact
contact.list_properties()

# Define variables which can be later used in a script
NEW_CONTACT_NAME = $new_contact_name  # Insert new contact name here
CONTACT_NEW_DESCRIPTION = $contact_new_description  #Insert new description for contact here

# alter contact
contact.alter(name=NEW_CONTACT_NAME)
contact.alter(description=CONTACT_NEW_DESCRIPTION)

# add new contact address to contact:
# 1. create a contact address
new_address = ContactAddress(
    name=CONTACT_ADDRESS_NAME,
    physical_address=PHYSICAL_ADDRESS,
    # the ContactDeliveryType values can be found in distribution_services/contact/contact.py
    delivery_type=ContactDeliveryType.EMAIL,
    device=DEVICE_ID,
    connection=conn
)
# 2. create new list of contact addresses
# combining list of existing contact addresses with a new one
new_address_list = contact.contact_addresses + [new_address]
# 3. update contact with a new list of contact addresses
contact.alter(contact_addresses=new_address_list)

# Define a variable which can be later used in a script
CONTACT_GROUP_NAME = $contact_group_name  # Insert name of contact group here

# get contact group
contact_group = ContactGroup(conn, name=CONTACT_GROUP_NAME)

# add contact to contact group
contact.add_to_contact_group(contact_group)

# remove contact from contact group
contact.remove_from_contact_group(contact_group)

# delete contact without a prompt
contact.delete(force=True)
